from typing import Optional
from unittest.mock import Mock

import pytest
from eth_typing import HexAddress

from raiden_contracts.constants import CONTRACTS_VERSION, EMPTY_ADDRESS, DeploymentModule
from raiden_contracts.contract_manager import (
    DeployedContract,
    contracts_data_path,
    contracts_deployed_path,
    get_contracts_deployment_info,
)
from raiden_contracts.deploy.contract_verifier import ContractVerifier
from raiden_contracts.utils.type_aliases import ChainID
from raiden_contracts.utils.versions import (
    contracts_version_provides_services,
    contracts_version_with_max_token_networks,
)


@pytest.mark.parametrize("version", [None, CONTRACTS_VERSION])
def test_deploy_data_dir_exists(version: Optional[str]) -> None:
    """ Make sure directories exist for deployment data """
    assert contracts_data_path(version).exists(), "deployment data do not exist"
    assert contracts_data_path(version).is_dir()


@pytest.mark.parametrize("version", [None, CONTRACTS_VERSION])
def test_deploy_data_dir_is_not_nested(version: Optional[str]) -> None:
    """ Make sure 'data' directories do not contain 'data*' recursively """
    assert list(contracts_data_path(version).glob("./data*")) == []


@pytest.mark.parametrize("version", [None, CONTRACTS_VERSION])
@pytest.mark.parametrize("chain_id", [3, 4, 5, 42])
@pytest.mark.parametrize("services", [False, True])
def test_deploy_data_file_exists(
    version: Optional[str], chain_id: ChainID, services: bool
) -> None:
    """ Make sure files exist for deployment data of each chain_id """
    assert contracts_deployed_path(chain_id, version, services).exists()


def reasonable_deployment_of_a_contract(deployed: DeployedContract) -> None:
    """ Checks an entry under deployment_*.json under a contract name """
    assert isinstance(deployed["address"], str)
    assert len(deployed["address"]) == 42
    assert isinstance(deployed["transaction_hash"], str)
    assert len(deployed["transaction_hash"]) == 66
    assert isinstance(deployed["block_number"], int)
    assert deployed["block_number"] > 0
    assert isinstance(deployed["gas_cost"], int)
    assert deployed["gas_cost"] > 0
    assert isinstance(deployed["constructor_arguments"], list)


RAIDEN_CONTRACT_NAMES = ("TokenNetworkRegistry", "SecretRegistry")


@pytest.mark.parametrize("version", [None])
@pytest.mark.parametrize("chain_id", [3, 4, 5, 42])
def test_deploy_data_has_fields_raiden(version: Optional[str], chain_id: ChainID) -> None:
    data = get_contracts_deployment_info(chain_id, version, module=DeploymentModule.RAIDEN)
    assert data
    assert data["contracts_version"] == version if version else CONTRACTS_VERSION
    assert data["chain_id"] == chain_id
    contracts = data["contracts"]
    for name in RAIDEN_CONTRACT_NAMES:
        deployed = contracts[name]
        reasonable_deployment_of_a_contract(deployed)


SERVICE_CONTRACT_NAMES = ("ServiceRegistry", "MonitoringService", "OneToN", "UserDeposit")


@pytest.mark.parametrize("version", [None])
@pytest.mark.parametrize("chain_id", [3, 4, 5, 42])
def test_deploy_data_has_fields_services(version: Optional[str], chain_id: ChainID) -> None:
    data = get_contracts_deployment_info(chain_id, version, module=DeploymentModule.SERVICES)
    assert data
    assert data["contracts_version"] == version if version else CONTRACTS_VERSION
    assert data["chain_id"] == chain_id
    contracts = data["contracts"]
    for name in SERVICE_CONTRACT_NAMES:
        deployed = contracts[name]
        reasonable_deployment_of_a_contract(deployed)


@pytest.mark.parametrize("version", [None])
@pytest.mark.parametrize("chain_id", [3, 4, 5, 42])
def test_deploy_data_all(version: Optional[str], chain_id: ChainID) -> None:
    data_all = get_contracts_deployment_info(chain_id, version, module=DeploymentModule.ALL)
    data_default = get_contracts_deployment_info(chain_id, version)
    assert data_all
    assert data_all == data_default

    for name in RAIDEN_CONTRACT_NAMES + SERVICE_CONTRACT_NAMES:
        deployed = data_all["contracts"][name]
        reasonable_deployment_of_a_contract(deployed)


def test_deploy_data_unknown_module() -> None:
    with pytest.raises(ValueError):
        get_contracts_deployment_info(3, None, module=None)  # type: ignore


def test_deploy_data_not_deployed() -> None:
    assert (
        get_contracts_deployment_info(ChainID(1), "0.8.0", module=DeploymentModule.RAIDEN) is None
    )


@pytest.mark.parametrize("chain_id", [3, 4, 42])
def test_deploy_data_for_redeyes_succeed(chain_id: ChainID) -> None:
    """ get_contracts_deployment_info() on RedEyes version should return a non-empty dictionary """
    assert get_contracts_deployment_info(chain_id, "0.4.0")


@pytest.mark.parametrize("chain_id", [3, 4, 5, 42])
def test_service_deploy_data_for_redeyes_fail(chain_id: ChainID) -> None:
    """ get_contracts_deployment_info() on RedEyes version should return a non-empty dictionary """
    with pytest.raises(ValueError):
        assert get_contracts_deployment_info(chain_id, "0.4.0", DeploymentModule.SERVICES)


def test_version_provides_services() -> None:
    assert not contracts_version_provides_services("0.3._")
    assert not contracts_version_provides_services("0.4.0")
    assert contracts_version_provides_services("0.8.0")
    assert contracts_version_provides_services("0.8.0_unlimited")
    assert contracts_version_provides_services("0.10.1")
    assert contracts_version_provides_services("0.11.0")
    with pytest.raises(ValueError):
        assert contracts_version_provides_services("not a semver")


def test_version_with_max_token_networks() -> None:
    assert contracts_version_with_max_token_networks(None)
    assert not contracts_version_with_max_token_networks("0.8.0_unlimited")


def test_verify_nonexistent_deployment(user_deposit_whole_balance_limit: int) -> None:
    """ Test verify_deployed_contracts_in_filesystem() with a non-existent deployment data. """
    web3_mock = Mock()
    web3_mock.version.network = 1
    # contracts_version 0.10.1 does not contain a main net deployment.
    verifier = ContractVerifier(web3=web3_mock, contracts_version="0.10.1")
    with pytest.raises(RuntimeError):
        verifier.verify_deployed_contracts_in_filesystem()
    with pytest.raises(RuntimeError):
        verifier.verify_deployed_service_contracts_in_filesystem(
            token_address=EMPTY_ADDRESS,
            user_deposit_whole_balance_limit=user_deposit_whole_balance_limit,
        )


def test_verify_existent_deployment() -> None:
    """ Test verify_deployed_contracts_in_filesystem() with an existent deployment data

    but with a fake web3 that returns a wrong block number for deployment.
    """
    web3_mock = Mock()
    web3_mock.version.network = 42
    web3_mock.eth.getTransactionReceipt = lambda _: {"blockNumber": 0}
    verifier = ContractVerifier(web3=web3_mock, contracts_version="0.11.1")
    # The Mock returns a wrong block number, so the comparison fails.
    with pytest.raises(RuntimeError):
        verifier.verify_deployed_contracts_in_filesystem()
    with pytest.raises(RuntimeError):
        verifier.verify_deployed_service_contracts_in_filesystem(
            token_address=HexAddress("0x3Aa761BcDB064179a1e37748D8A5F577a177Be5c"),
            user_deposit_whole_balance_limit=2 ** 256 - 1,
        )


def test_verify_existent_deployment_with_wrong_code() -> None:
    """ Test verify_deployed_contracts_in_filesystem() with an existent deployment data

    but with a fake web3 that does not return the correct code.
    """
    web3_mock = Mock()
    web3_mock.version.network = 42
    web3_mock.eth.getTransactionReceipt = lambda _: {
        "blockNumber": 10711807,
        "gasUsed": 555366,
        "contractAddress": "0x8Ff327f7ed03cD6Bd5e611E9e404B47d8c9Db81E",
    }
    verifier = ContractVerifier(web3=web3_mock, contracts_version="0.11.1")
    # The Mock returns a wrong block number, so the comparison fails.
    with pytest.raises(RuntimeError):
        verifier.verify_deployed_contracts_in_filesystem()
    with pytest.raises(RuntimeError):
        verifier.verify_deployed_service_contracts_in_filesystem(
            token_address=HexAddress("0x3Aa761BcDB064179a1e37748D8A5F577a177Be5c"),
            user_deposit_whole_balance_limit=2 ** 256 - 1,
        )



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3.contract import Contract
from web3.exceptions import ValidationError

from raiden_contracts.constants import EMPTY_ADDRESS, ChannelEvent
from raiden_contracts.utils.events import check_channel_settled


@pytest.mark.skip(reason="Delayed until another milestone")
def test_cooperative_settle_channel_call(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_cooperative_settle_signatures: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    balance_A = 5
    balance_B = 25

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (signature_A, signature_B) = create_cooperative_settle_signatures(
        [A, B], channel_identifier, A, balance_A, B, balance_B
    )

    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier, A, -1, B, balance_B, signature_A, signature_B
        )
    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier, A, balance_A, B, -1, signature_A, signature_B
        )
    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier, 0x0, balance_A, B, balance_B, signature_A, signature_B
        )
    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier, A, balance_A, 0x0, balance_B, signature_A, signature_B
        )
    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier, A, balance_A, B, balance_B, 0x0, signature_B
        )
    with pytest.raises(ValidationError):
        token_network.functions.cooperativeSettle(
            channel_identifier, A, balance_A, B, balance_B, signature_A, 0x0
        )

    with pytest.raises(TransactionFailed):
        token_network.functions.cooperativeSettle(
            channel_identifier, EMPTY_ADDRESS, balance_A, B, balance_B, signature_A, signature_B
        ).call_and_transact({"from": C})
    with pytest.raises(TransactionFailed):
        token_network.functions.cooperativeSettle(
            channel_identifier, A, balance_A, EMPTY_ADDRESS, balance_B, signature_A, signature_B
        ).call_and_transact({"from": C})

    token_network.functions.cooperativeSettle(
        channel_identifier, A, balance_A, B, balance_B, signature_A, signature_B
    ).call_and_transact({"from": C})


@pytest.mark.skip(reason="Delayed until another milestone")
def test_cooperative_settle_channel_signatures(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_cooperative_settle_signatures: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    balance_A = 4
    balance_B = 26

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (signature_A, signature_B, signature_C) = create_cooperative_settle_signatures(
        [A, B, C], channel_identifier, A, balance_A, B, balance_B
    )

    with pytest.raises(TransactionFailed):
        token_network.functions.cooperativeSettle(
            channel_identifier, A, balance_A, B, balance_B, signature_C, signature_B
        ).call({"from": C})
    with pytest.raises(TransactionFailed):
        token_network.functions.cooperativeSettle(
            channel_identifier, A, balance_A, B, balance_B, signature_A, signature_C
        ).call({"from": C})
    with pytest.raises(TransactionFailed):
        token_network.functions.cooperativeSettle(
            channel_identifier, A, balance_B, B, balance_A, signature_A, signature_B
        ).call({"from": C})

    token_network.functions.cooperativeSettle(
        channel_identifier, A, balance_A, B, balance_B, signature_A, signature_B
    ).call_and_transact({"from": C})


@pytest.mark.skip(reason="Delayed until another milestone")
def test_cooperative_settle_channel_0(
    custom_token: Contract,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_cooperative_settle_signatures: Callable,
    cooperative_settle_state_tests: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    balance_A = 0
    balance_B = 30

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (signature_A, signature_B, _) = create_cooperative_settle_signatures(
        [A, B, C], channel_identifier, A, balance_A, B, balance_B
    )

    pre_account_balance_A = custom_token.functions.balanceOf(A).call()
    pre_account_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    token_network.functions.cooperativeSettle(
        channel_identifier, A, balance_A, B, balance_B, signature_A, signature_B
    ).call_and_transact({"from": C})

    cooperative_settle_state_tests(
        channel_identifier,
        A,
        balance_A,
        B,
        balance_B,
        pre_account_balance_A,
        pre_account_balance_B,
        pre_balance_contract,
    )


@pytest.mark.skip(reason="Delayed until another milestone")
def test_cooperative_settle_channel_00(
    custom_token: Contract,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_cooperative_settle_signatures: Callable,
    cooperative_settle_state_tests: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 0
    deposit_B = 0
    balance_A = 0
    balance_B = 0

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (signature_A, signature_B, _) = create_cooperative_settle_signatures(
        [A, B, C], channel_identifier, A, balance_A, B, balance_B
    )

    pre_account_balance_A = custom_token.functions.balanceOf(A).call()
    pre_account_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    token_network.functions.cooperativeSettle(
        channel_identifier, A, balance_A, B, balance_B, signature_A, signature_B
    ).call_and_transact({"from": C})

    cooperative_settle_state_tests(
        channel_identifier,
        A,
        balance_A,
        B,
        balance_B,
        pre_account_balance_A,
        pre_account_balance_B,
        pre_balance_contract,
    )


@pytest.mark.skip(reason="Delayed until another milestone")
def test_cooperative_settle_channel_state(
    custom_token: Contract,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_cooperative_settle_signatures: Callable,
    cooperative_settle_state_tests: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    balance_A = 5
    balance_B = 25
    assert deposit_A + deposit_B == balance_A + balance_B

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (signature_A, signature_B) = create_cooperative_settle_signatures(
        [A, B], channel_identifier, A, balance_A, B, balance_B
    )

    pre_account_balance_A = custom_token.functions.balanceOf(A).call()
    pre_account_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    token_network.functions.cooperativeSettle(
        channel_identifier, A, balance_A, B, balance_B, signature_A, signature_B
    ).call_and_transact({"from": C})

    cooperative_settle_state_tests(
        channel_identifier,
        A,
        balance_A,
        B,
        balance_B,
        pre_account_balance_A,
        pre_account_balance_B,
        pre_balance_contract,
    )


@pytest.mark.skip(reason="Delayed until another milestone")
def test_cooperative_settle_channel_state_withdraw(
    custom_token: Contract,
    token_network: Contract,
    create_channel_and_deposit: Callable,
    withdraw_channel: Callable,
    get_accounts: Callable,
    create_cooperative_settle_signatures: Callable,
    cooperative_settle_state_tests: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    withdraw_A = 3
    withdraw_B = 7
    balance_A = 5
    balance_B = 15
    assert deposit_A + deposit_B == withdraw_A + withdraw_B + balance_A + balance_B

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)
    withdraw_channel(channel_identifier, A, withdraw_A, B)
    withdraw_channel(channel_identifier, B, withdraw_B, A)

    (signature_A, signature_B) = create_cooperative_settle_signatures(
        [A, B], channel_identifier, A, balance_A, B, balance_B
    )

    pre_account_balance_A = custom_token.functions.balanceOf(A).call()
    pre_account_balance_B = custom_token.functions.balanceOf(B).call()
    pre_balance_contract = custom_token.functions.balanceOf(token_network.address).call()

    token_network.functions.cooperativeSettle(
        channel_identifier, A, balance_A, B, balance_B, signature_A, signature_B
    ).call_and_transact({"from": C})

    cooperative_settle_state_tests(
        channel_identifier,
        A,
        balance_A,
        B,
        balance_B,
        pre_account_balance_A,
        pre_account_balance_B,
        pre_balance_contract,
    )


@pytest.mark.skip(reason="Delayed until another milestone")
def test_cooperative_settle_channel_bigger_withdraw(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    withdraw_channel: Callable,
    get_accounts: Callable,
    create_cooperative_settle_signatures: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    withdraw_A = 3
    withdraw_B = 7
    balance_A = 6
    balance_B = 15
    assert deposit_A + deposit_B < withdraw_A + withdraw_B + balance_A + balance_B

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)
    withdraw_channel(channel_identifier, A, withdraw_A, B)
    withdraw_channel(channel_identifier, B, withdraw_B, A)

    (signature_A, signature_B) = create_cooperative_settle_signatures(
        [A, B], channel_identifier, A, balance_A, B, balance_B
    )

    with pytest.raises(TransactionFailed):
        token_network.functions.cooperativeSettle(
            channel_identifier, A, balance_A, B, balance_B, signature_A, signature_B
        ).call({"from": C})


@pytest.mark.skip(reason="Delayed until another milestone")
def test_cooperative_settle_channel_wrong_balances(
    token_network: Contract,
    create_channel_and_deposit: Callable,
    get_accounts: Callable,
    create_cooperative_settle_signatures: Callable,
) -> None:
    (A, B, C) = get_accounts(3)
    deposit_A = 20
    deposit_B = 10
    balance_A = 7
    balance_B = 23

    balance_A_fail1 = 20
    balance_B_fail1 = 11
    balance_A_fail2 = 6
    balance_B_fail2 = 8

    channel_identifier = create_channel_and_deposit(A, B, deposit_A, deposit_B)

    (signature_A, signature_B) = create_cooperative_settle_signatures(
        [A, B], channel_identifier, A, balance_A, B, balance_B
    )
    (signature_A_fail1, signature_B_fail1) = create_cooperative_settle_signatures(
        [A, B], channel_identifier, A, balance_A_fail1, B, balance_B_fail1
    )
    (signature_A_fail2, signature_B_fail2) = create_cooperative_settle_signatures(
        [A, B], channel_identifier, A, balance_A_fail2, B, balance_B_fail2
    )

    with pytest.raises(TransactionFailed):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            A,
            balance_A_fail1,
            B,
            balance_B_fail1,
            signature_A_fail1,
            signature_B_fail1,
        ).call({"from": C})
    with pytest.raises(TransactionFailed):
        token_network.functions.cooperativeSettle(
            channel_identifier,
            A,
            balance_A_fail2,
            B,
            balance_B_fail2,
            signature_A_fail2,
            signature_B_fail2,
        ).call({"from": C})

    token_network.functions.cooperativeSettle(
        channel_identifier, A, balance_A, B, balance_B, signature_A, signature_B
    ).call_and_transact({"from": C})


@pytest.mark.skip(reason="Delayed until another milestone")
def test_cooperative_close_replay_reopened_channel(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    create_cooperative_settle_signatures: Callable,
) -> None:
    (A, B) = get_accounts(2)
    deposit_A = 15
    deposit_B = 10
    balance_A = 2
    balance_B = 23

    channel_identifier1 = create_channel(A, B)[0]
    channel_deposit(channel_identifier1, A, deposit_A, B)
    channel_deposit(channel_identifier1, B, deposit_B, A)

    (signature_A, signature_B) = create_cooperative_settle_signatures(
        [A, B], channel_identifier1, B, balance_B, A, balance_A
    )

    token_network.functions.cooperativeSettle(
        channel_identifier1, B, balance_B, A, balance_A, signature_B, signature_A
    ).call_and_transact({"from": B})

    # Reopen the channel and make sure we cannot use the old balance proof
    channel_identifier2 = create_channel(A, B)[0]
    channel_deposit(channel_identifier2, A, deposit_A, B)
    channel_deposit(channel_identifier2, B, deposit_B, A)

    assert channel_identifier1 != channel_identifier2
    with pytest.raises(TransactionFailed):
        token_network.functions.cooperativeSettle(
            channel_identifier2, B, balance_B, A, balance_A, signature_B, signature_A
        ).call({"from": B})

    # Signed message with the correct channel identifier must work
    (signature_A2, signature_B2) = create_cooperative_settle_signatures(
        [A, B], channel_identifier2, B, balance_B, A, balance_A
    )
    token_network.functions.cooperativeSettle(
        channel_identifier2, B, balance_B, A, balance_A, signature_B2, signature_A2
    ).call_and_transact({"from": B})


@pytest.mark.skip(reason="Delayed until another milestone")
def test_cooperative_settle_channel_event(
    get_accounts: Callable,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    create_cooperative_settle_signatures: Callable,
    event_handler: Callable,
) -> None:
    ev_handler = event_handler(token_network)
    (A, B) = get_accounts(2)
    deposit_A = 10
    balance_A = 2
    balance_B = 8
    channel_identifier = create_channel(A, B)[0]
    channel_deposit(channel_identifier, A, deposit_A, B)

    (signature_A, signature_B) = create_cooperative_settle_signatures(
        [A, B], channel_identifier, B, balance_B, A, balance_A
    )

    txn_hash = token_network.functions.cooperativeSettle(
        channel_identifier, B, balance_B, A, balance_A, signature_B, signature_A
    ).call_and_transact({"from": B})

    ev_handler.add(
        txn_hash,
        ChannelEvent.SETTLED,
        check_channel_settled(channel_identifier, balance_B, balance_A),
    )
    ev_handler.check()



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import OneToNEvent
from raiden_contracts.utils.proofs import sign_one_to_n_iou


def test_claim(
    user_deposit_contract: Contract,
    one_to_n_contract: Contract,
    deposit_to_udc: Callable,
    get_accounts: Callable,
    get_private_key: Callable,
    web3: Web3,
    event_handler: Callable,
) -> None:
    ev_handler = event_handler(one_to_n_contract)
    (A, B) = get_accounts(2)
    deposit_to_udc(A, 30)

    # happy case
    amount = 10
    expiration = web3.eth.blockNumber + 2
    chain_id = int(web3.version.network)

    # IOU expired
    with pytest.raises(TransactionFailed):
        bad_expiration = web3.eth.blockNumber - 1
        signature = sign_one_to_n_iou(
            get_private_key(A),
            sender=A,
            receiver=B,
            amount=amount,
            expiration_block=bad_expiration,
            one_to_n_address=one_to_n_contract.address,
            chain_id=chain_id,
        )
        one_to_n_contract.functions.claim(
            A, B, amount, bad_expiration, one_to_n_contract.address, signature
        ).call({"from": A})

    # Wrong OneToN address
    with pytest.raises(TransactionFailed):
        signature = sign_one_to_n_iou(
            get_private_key(A),
            sender=A,
            receiver=B,
            amount=amount,
            expiration_block=expiration,
            one_to_n_address=A,  # Inject an error
            chain_id=chain_id,
        )
        one_to_n_contract.functions.claim(A, B, amount, expiration, A, signature).call({"from": A})

    # Wrong chain_id
    with pytest.raises(TransactionFailed):
        signature = sign_one_to_n_iou(
            get_private_key(A),
            sender=A,
            receiver=B,
            amount=amount,
            expiration_block=expiration,
            one_to_n_address=one_to_n_contract.address,
            chain_id=chain_id + 2,  # Inject an error
        )
        one_to_n_contract.functions.claim(A, B, amount, expiration, A, signature).call({"from": A})

    # bad signature
    with pytest.raises(TransactionFailed):
        expiration = web3.eth.blockNumber + 1
        signature = sign_one_to_n_iou(
            get_private_key(A),
            sender=A,
            receiver=B,
            amount=amount + 1,  # this does not match amount below
            expiration_block=expiration,
            one_to_n_address=one_to_n_contract.address,
            chain_id=chain_id,
        )
        one_to_n_contract.functions.claim(
            A, B, amount, expiration, one_to_n_contract.address, signature
        ).call({"from": A})

    signature = sign_one_to_n_iou(
        get_private_key(A),
        sender=A,
        receiver=B,
        amount=amount,
        expiration_block=expiration,
        one_to_n_address=one_to_n_contract.address,
        chain_id=chain_id,
    )

    tx_hash = one_to_n_contract.functions.claim(
        sender=A,
        receiver=B,
        amount=amount,
        expiration_block=expiration,
        one_to_n_address=one_to_n_contract.address,
        signature=signature,
    ).call_and_transact({"from": A})

    ev_handler.assert_event(
        tx_hash,
        OneToNEvent.CLAIMED,
        dict(sender=A, receiver=B, expiration_block=expiration, transferred=amount),
    )
    assert user_deposit_contract.functions.balances(A).call() == 20
    assert user_deposit_contract.functions.balances(B).call() == 10

    # can't be claimed twice
    with pytest.raises(TransactionFailed):
        one_to_n_contract.functions.claim(
            A, B, amount, expiration, one_to_n_contract.address, signature
        ).call({"from": A})


def test_claim_with_insufficient_deposit(
    user_deposit_contract: Contract,
    one_to_n_contract: Contract,
    deposit_to_udc: Callable,
    get_accounts: Callable,
    get_private_key: Callable,
    web3: Web3,
    event_handler: Callable,
) -> None:
    ev_handler = event_handler(one_to_n_contract)
    (A, B) = get_accounts(2)
    deposit_to_udc(A, 6)
    chain_id = int(web3.version.network)

    amount = 10
    expiration = web3.eth.blockNumber + 1
    signature = sign_one_to_n_iou(
        get_private_key(A),
        sender=A,
        receiver=B,
        amount=amount,
        expiration_block=expiration,
        one_to_n_address=one_to_n_contract.address,
        chain_id=chain_id,
    )

    # amount is 10, but only 6 are in deposit
    # check return value (transactions don't give back return values, so use call)
    assert (
        one_to_n_contract.functions.claim(
            A, B, amount, expiration, one_to_n_contract.address, signature
        ).call({"from": A})
        == 6
    )
    # check that transaction succeeds
    one_to_n_contract.functions.claim(
        A, B, amount, expiration, one_to_n_contract.address, signature
    ).call_and_transact({"from": A})

    assert user_deposit_contract.functions.balances(A).call() == 0
    assert user_deposit_contract.functions.balances(B).call() == 6

    # claim can be retried when transferred amount was 0
    expiration = web3.eth.blockNumber + 10
    signature = sign_one_to_n_iou(
        get_private_key(A),
        sender=A,
        receiver=B,
        amount=amount,
        expiration_block=expiration,
        one_to_n_address=one_to_n_contract.address,
        chain_id=chain_id,
    )
    one_to_n_contract.functions.claim(
        A, B, amount, expiration, one_to_n_contract.address, signature
    ).call_and_transact({"from": A})
    deposit_to_udc(A, 6 + 4)
    tx_hash = one_to_n_contract.functions.claim(
        A, B, amount, expiration, one_to_n_contract.address, signature
    ).call_and_transact({"from": A})
    ev_handler.assert_event(
        tx_hash,
        OneToNEvent.CLAIMED,
        dict(sender=A, receiver=B, expiration_block=expiration, transferred=4),
    )



import logging
from typing import Callable, List, Optional

import pytest
from eth_tester.exceptions import TransactionFailed
from eth_typing import HexAddress
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.contract_manager import ContractManager
from raiden_contracts.tests.utils.constants import CONTRACT_DEPLOYER_ADDRESS

log = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def deploy_tester_contract(
    web3: Web3, contracts_manager: ContractManager, deploy_contract: Callable
) -> Callable:
    """Returns a function that can be used to deploy a named contract,
    using conract manager to compile the bytecode and get the ABI"""

    def f(contract_name: str, args: Optional[List] = None) -> Contract:
        json_contract = contracts_manager.get_contract(contract_name)
        contract = deploy_contract(
            web3, CONTRACT_DEPLOYER_ADDRESS, json_contract["abi"], json_contract["bin"], args
        )
        return contract

    return f


@pytest.fixture(scope="session")
def deploy_contract_txhash() -> Callable[..., str]:
    """Returns a function that deploys a compiled contract, returning a txhash"""

    def fn(
        web3: Web3, deployer_address: HexAddress, abi: List, bytecode: str, args: Optional[List]
    ) -> str:
        if args is None:
            args = []
        contract = web3.eth.contract(abi=abi, bytecode=bytecode)
        # Failure does not fire an exception.  Check the receipt for status.
        return contract.constructor(*args).transact({"from": deployer_address})

    return fn


@pytest.fixture(scope="session")
def deploy_contract(deploy_contract_txhash: Callable) -> Callable:
    """Returns a function that deploys a compiled contract"""

    def fn(
        web3: Web3, deployer_address: HexAddress, abi: List, bytecode: str, args: List
    ) -> Contract:
        contract = web3.eth.contract(abi=abi, bytecode=bytecode)
        txhash = deploy_contract_txhash(web3, deployer_address, abi, bytecode, args)
        contract_address = web3.eth.getTransactionReceipt(txhash).contractAddress
        web3.testing.mine(1)

        if web3.eth.getTransactionReceipt(txhash).status != 1:
            raise TransactionFailed("deployment failed")

        return contract(contract_address)

    return fn


@pytest.fixture
def deploy_tester_contract_txhash(
    web3: Web3, contracts_manager: ContractManager, deploy_contract_txhash: Callable
) -> Callable:
    """Returns a function that can be used to deploy a named contract,
    but returning txhash only"""

    def f(contract_name: str, args: Optional[List] = None) -> str:
        json_contract = contracts_manager.get_contract(contract_name)
        txhash = deploy_contract_txhash(
            web3, CONTRACT_DEPLOYER_ADDRESS, json_contract["abi"], json_contract["bin"], args
        )
        return txhash

    return f


@pytest.fixture
def utils_contract(deploy_tester_contract: Callable) -> Contract:
    """Deployed Utils contract"""
    return deploy_tester_contract("Utils")



from unittest.mock import Mock

import pytest

from raiden_contracts.utils.transaction import check_successful_tx


def test_check_successful_tx_with_status_zero() -> None:
    web3_mock = Mock()
    web3_mock.eth.getTransactionReceipt.return_value = {"blockNumber": 300, "status": 0}
    txid = "abcdef"
    with pytest.raises(ValueError):
        check_successful_tx(web3=web3_mock, txid=txid)
    web3_mock.eth.getTransactionReceipt.assert_called_with(txid)
    web3_mock.eth.getTransaction.assert_called_with(txid)


def test_check_successful_tx_with_nonexistent_status() -> None:
    """ check_successful_tx() with a receipt without status field should raise a KeyError """
    web3_mock = Mock()
    web3_mock.eth.getTransactionReceipt.return_value = {"blockNumber": 300}
    txid = "abcdef"
    with pytest.raises(KeyError):
        check_successful_tx(web3=web3_mock, txid=txid)
    web3_mock.eth.getTransactionReceipt.assert_called_with(txid)
    web3_mock.eth.getTransaction.assert_called_with(txid)


def test_check_successful_tx_with_gas_completely_used() -> None:
    web3_mock = Mock()
    gas = 30000
    web3_mock.eth.getTransactionReceipt.return_value = {
        "blockNumber": 300,
        "status": 1,
        "gasUsed": gas,
    }
    web3_mock.eth.getTransaction.return_value = {"gas": gas}
    txid = "abcdef"
    with pytest.raises(ValueError):
        check_successful_tx(web3=web3_mock, txid=txid)
    web3_mock.eth.getTransactionReceipt.assert_called_with(txid)
    web3_mock.eth.getTransaction.assert_called_with(txid)


def test_check_successful_tx_successful_case() -> None:
    web3_mock = Mock()
    gas = 30000
    receipt = {"blockNumber": 300, "status": 1, "gasUsed": gas - 10}
    web3_mock.eth.getTransactionReceipt.return_value = receipt
    txinfo = {"gas": gas}
    web3_mock.eth.getTransaction.return_value = txinfo
    txid = "abcdef"
    assert check_successful_tx(web3=web3_mock, txid=txid) == (receipt, txinfo)
    web3_mock.eth.getTransactionReceipt.assert_called_with(txid)
    web3_mock.eth.getTransaction.assert_called_with(txid)



from typing import Callable, Dict, List

import pytest
from eth_abi import encode_single
from eth_tester.exceptions import TransactionFailed
from eth_typing import HexAddress
from eth_utils import to_checksum_address
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import (
    LOCKSROOT_OF_NO_LOCKS,
    TEST_SETTLE_TIMEOUT_MIN,
    MonitoringServiceEvent,
)
from raiden_contracts.tests.utils import SERVICE_DEPOSIT
from raiden_contracts.utils.proofs import sign_reward_proof

REWARD_AMOUNT = 10


@pytest.fixture
def ms_address(
    get_accounts: Callable, custom_token: Contract, service_registry: Contract
) -> HexAddress:
    (ms,) = get_accounts(1)

    # register MS in the ServiceRegistry contract
    custom_token.functions.mint(2 * SERVICE_DEPOSIT).call_and_transact({"from": ms})
    custom_token.functions.approve(service_registry.address, SERVICE_DEPOSIT).call_and_transact(
        {"from": ms}
    )
    service_registry.functions.deposit(SERVICE_DEPOSIT).call_and_transact({"from": ms})

    return ms


@pytest.fixture
def setup_monitor_data(
    get_accounts: Callable,
    deposit_to_udc: Callable,
    create_channel: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    token_network: Contract,
    ms_address: HexAddress,
    get_private_key: Callable,
) -> Callable:
    def f(monitoring_service_contract: Contract) -> Dict:
        # Create two parties and a channel between them
        (A, B) = get_accounts(2, privkeys=["0x" + "1" * 64, "0x" + "2" * 64])
        deposit_to_udc(B, REWARD_AMOUNT)
        channel_identifier = create_channel(A, B)[0]

        # Create balance proofs
        balance_proof_A = create_balance_proof(
            channel_identifier, B, transferred_amount=10, nonce=1
        )
        balance_proof_B = create_balance_proof(
            channel_identifier, A, transferred_amount=20, nonce=2
        )

        # Add signatures by non_closing_participant
        closing_signature_A = create_balance_proof_countersignature(
            A, channel_identifier, *balance_proof_A
        )
        non_closing_signature_B = create_balance_proof_countersignature(
            B, channel_identifier, *balance_proof_B
        )
        reward_proof_signature = sign_reward_proof(
            privatekey=get_private_key(B),
            monitoring_service_contract_address=monitoring_service_contract.address,
            chain_id=token_network.functions.chain_id().call(),
            non_closing_signature=non_closing_signature_B,
            reward_amount=REWARD_AMOUNT,
        )

        # close channel
        token_network.functions.closeChannel(
            channel_identifier, B, A, *balance_proof_A, closing_signature_A
        ).call_and_transact({"from": A})

        # calculate when this MS is allowed to monitor
        (settle_block_number, _) = token_network.functions.getChannelInfo(
            channel_identifier, A, B
        ).call()
        first_allowed = monitoring_service_contract.functions.firstBlockAllowedToMonitor(
            closed_at_block=settle_block_number - TEST_SETTLE_TIMEOUT_MIN,
            settle_timeout=TEST_SETTLE_TIMEOUT_MIN,
            participant1=A,
            participant2=B,
            monitoring_service_address=ms_address,
        ).call()

        # return args for `monitor` function
        return {
            "participants": (A, B),
            "balance_proof_A": balance_proof_A,
            "balance_proof_B": balance_proof_B,
            "non_closing_signature": non_closing_signature_B,
            "reward_proof_signature": reward_proof_signature,
            "channel_identifier": channel_identifier,
            "first_allowed": first_allowed,
        }

    return f


@pytest.fixture
def monitor_data(setup_monitor_data: Callable, monitoring_service_external: Contract) -> Dict:
    return setup_monitor_data(monitoring_service_external)


@pytest.fixture
def monitor_data_internal(
    setup_monitor_data: Callable, monitoring_service_internals: Contract
) -> Dict:
    return setup_monitor_data(monitoring_service_internals)


@pytest.mark.parametrize("with_settle", [True, False])
def test_claimReward_with_settle_call(
    token_network: Contract,
    monitoring_service_external: Contract,
    user_deposit_contract: Contract,
    event_handler: Callable,
    monitor_data: Dict,
    ms_address: HexAddress,
    web3: Web3,
    with_settle: bool,
) -> None:
    A, B = monitor_data["participants"]
    channel_identifier = monitor_data["channel_identifier"]

    # wait until MS is allowed to monitor
    token_network.web3.testing.mine(monitor_data["first_allowed"] - web3.eth.blockNumber)

    # MS updates closed channel on behalf of B
    txn_hash = monitoring_service_external.functions.monitor(
        A,
        B,
        *monitor_data["balance_proof_B"],
        monitor_data["non_closing_signature"],
        REWARD_AMOUNT,
        token_network.address,
        monitor_data["reward_proof_signature"],
    ).call_and_transact({"from": ms_address})

    # claiming before settlement timeout must fail
    with pytest.raises(TransactionFailed):
        monitoring_service_external.functions.claimReward(
            channel_identifier, token_network.address, A, B
        ).call({"from": ms_address})

    # Settle channel after settle_timeout elapsed
    token_network.web3.testing.mine(4)
    if with_settle:
        token_network.functions.settleChannel(
            channel_identifier,
            B,  # participant_B
            10,  # participant_B_transferred_amount
            0,  # participant_B_locked_amount
            LOCKSROOT_OF_NO_LOCKS,  # participant_B_locksroot
            A,  # participant_A
            20,  # participant_A_transferred_amount
            0,  # participant_A_locked_amount
            LOCKSROOT_OF_NO_LOCKS,  # participant_A_locksroot
        ).call_and_transact()

    # Claim reward for MS
    monitoring_service_external.functions.claimReward(
        channel_identifier, token_network.address, A, B
    ).call_and_transact({"from": ms_address})

    # Check REWARD_CLAIMED event
    reward_identifier = Web3.sha3(
        encode_single("uint256", channel_identifier) + Web3.toBytes(hexstr=token_network.address)
    )
    ms_ev_handler = event_handler(monitoring_service_external)
    ms_ev_handler.assert_event(
        txn_hash,
        MonitoringServiceEvent.REWARD_CLAIMED,
        dict(ms_address=ms_address, amount=REWARD_AMOUNT, reward_identifier=reward_identifier),
    )

    # Check that MS balance has increased by claiming the reward
    ms_balance_after_reward = user_deposit_contract.functions.balances(ms_address).call()
    assert ms_balance_after_reward == REWARD_AMOUNT


def test_monitor(
    token_network: Contract,
    monitoring_service_external: Contract,
    monitor_data: Dict,
    ms_address: HexAddress,
    event_handler: Callable,
    web3: Web3,
) -> None:
    A, B = monitor_data["participants"]

    # UpdateNonClosingBalanceProof is tested speparately, so we assume that all
    # parameters passed to it are handled correctly.

    # changing reward amount must lead to a failure during reward signature check
    with pytest.raises(TransactionFailed):
        txn_hash = monitoring_service_external.functions.monitor(
            A,
            B,
            *monitor_data["balance_proof_B"],
            monitor_data["non_closing_signature"],
            REWARD_AMOUNT + 1,
            token_network.address,
            monitor_data["reward_proof_signature"],
        ).call({"from": ms_address})

    # only registered service provicers may call `monitor`
    with pytest.raises(TransactionFailed):
        txn_hash = monitoring_service_external.functions.monitor(
            A,
            B,
            *monitor_data["balance_proof_B"],
            monitor_data["non_closing_signature"],
            REWARD_AMOUNT + 1,
            token_network.address,
            monitor_data["reward_proof_signature"],
        ).call({"from": B})

    # monitoring too early must fail
    with pytest.raises(TransactionFailed):
        assert web3.eth.blockNumber < monitor_data["first_allowed"]
        txn_hash = monitoring_service_external.functions.monitor(
            A,
            B,
            *monitor_data["balance_proof_B"],
            monitor_data["non_closing_signature"],
            REWARD_AMOUNT,
            token_network.address,
            monitor_data["reward_proof_signature"],
        ).call_and_transact({"from": ms_address})

    # wait until MS is allowed to monitor
    token_network.web3.testing.mine(monitor_data["first_allowed"] - web3.eth.blockNumber)

    # successful monitor call
    txn_hash = monitoring_service_external.functions.monitor(
        A,
        B,
        *monitor_data["balance_proof_B"],
        monitor_data["non_closing_signature"],
        REWARD_AMOUNT,
        token_network.address,
        monitor_data["reward_proof_signature"],
    ).call_and_transact({"from": ms_address})

    # NEW_BALANCE_PROOF_RECEIVED must get emitted
    ms_ev_handler = event_handler(monitoring_service_external)
    ms_ev_handler.assert_event(
        txn_hash,
        MonitoringServiceEvent.NEW_BALANCE_PROOF_RECEIVED,
        dict(
            token_network_address=token_network.address,
            channel_identifier=monitor_data["channel_identifier"],
            reward_amount=REWARD_AMOUNT,
            nonce=monitor_data["balance_proof_B"][1],
            ms_address=ms_address,
            raiden_node_address=B,
        ),
    )


def test_updateReward(
    monitoring_service_internals: Contract,
    ms_address: HexAddress,
    token_network: Contract,
    monitor_data_internal: Dict,
) -> None:
    A, B = monitor_data_internal["participants"]
    reward_identifier = Web3.sha3(
        encode_single("uint256", monitor_data_internal["channel_identifier"])
        + Web3.toBytes(hexstr=token_network.address)
    )

    def update_with_nonce(nonce: int) -> None:
        monitoring_service_internals.functions.updateRewardPublic(
            token_network.address,
            A,
            B,
            REWARD_AMOUNT,
            nonce,
            ms_address,
            monitor_data_internal["non_closing_signature"],
            monitor_data_internal["reward_proof_signature"],
        ).call_and_transact({"from": ms_address})

    # normal first call succeeds
    update_with_nonce(2)
    assert monitoring_service_internals.functions.rewardNonce(reward_identifier).call() == 2

    # calling again with same nonce fails
    with pytest.raises(TransactionFailed):
        update_with_nonce(2)

    # calling again with higher nonce succeeds
    update_with_nonce(3)
    assert monitoring_service_internals.functions.rewardNonce(reward_identifier).call() == 3


def test_firstAllowedBlock(monitoring_service_external: Contract) -> None:
    def call(
        addresses: List[HexAddress], closed_at_block: int = 1000, settle_timeout: int = 100
    ) -> int:
        first_allowed = monitoring_service_external.functions.firstBlockAllowedToMonitor(
            closed_at_block=closed_at_block,
            settle_timeout=settle_timeout,
            participant1=to_checksum_address("0x%040x" % addresses[0]),
            participant2=to_checksum_address("0x%040x" % addresses[1]),
            monitoring_service_address=to_checksum_address("0x%040x" % addresses[2]),
        ).call()
        assert closed_at_block < first_allowed <= closed_at_block + settle_timeout
        return first_allowed

    # Basic example
    assert call([1, 2, 3]) == 1000 + 30 + (1 + 2 + 3)

    # Modulo used for one address
    assert call([100, 2, 3]) == 1000 + 30 + (100 % 50 + 2 + 3)

    # Modulo not used for any single address, but the sum of them
    assert call([40, 40, 40]) == 1000 + 30 + (40 + 40 + 40) % 50

    # Show that high address values don't cause overflows
    MAX_ADDRESS = 256 ** 20 - 1
    assert call([MAX_ADDRESS] * 3) == 1000 + 30 + (3 * MAX_ADDRESS) % 50

    # The highest settle_timeout does not cause overflows
    MAX_SETTLE_TIMEOUT = (2 ** 256 - 1) // 100 - 1
    assert call([1, 2, 3], settle_timeout=MAX_SETTLE_TIMEOUT) == 1000 + (
        MAX_SETTLE_TIMEOUT * 30
    ) // 100 + (1 + 2 + 3)

    # Extreme settle_timeout that would cause overflows will make the
    # transaction fail instead of giving the wrong result
    with pytest.raises(TransactionFailed):
        assert call([1, 2, 3], settle_timeout=MAX_SETTLE_TIMEOUT + 1)


def test_recoverAddressFromRewardProof(
    monitor_data_internal: Dict, token_network: Contract, monitoring_service_internals: Contract
) -> None:
    _, B = monitor_data_internal["participants"]

    recoverAddressFromRewardProof = (
        monitoring_service_internals.functions.recoverAddressFromRewardProofPublic
    )
    recovered_address = recoverAddressFromRewardProof(
        monitoring_service_contract_address=monitoring_service_internals.address,
        chain_id=token_network.functions.chain_id().call(),
        non_closing_signature=monitor_data_internal["non_closing_signature"],
        reward_amount=REWARD_AMOUNT,
        signature=monitor_data_internal["reward_proof_signature"],
    ).call()
    assert recovered_address == B



"""ContractSourceManager knows the sources and how to compile them."""
import hashlib
import json
from os import chdir
from pathlib import Path
from typing import Dict

from solc import compile_files

from raiden_contracts.constants import PRECOMPILED_DATA_FIELDS, DeploymentModule
from raiden_contracts.contract_manager import ContractManager

_BASE = Path(__file__).parent


class ContractSourceManagerCompilationError(RuntimeError):
    """Compilation failed for infrastructural reasons (lack of the compiler,
    failure to take checksums)."""


class ContractSourceManagerVerificationError(RuntimeError):
    """Failure in comparing contracts.json contents against sources."""


class ContractSourceManager:
    """ ContractSourceManager knows how to compile contracts """

    def __init__(self, path: Dict[str, Path]) -> None:
        """ Parmas: a dictionary of directories which contain solidity files to compile """
        if not isinstance(path, dict):
            raise TypeError("Wrong type of argument given for ContractSourceManager()")
        self.contracts_source_dirs = path

    def _compile_all_contracts(self) -> Dict:
        """
        Compile solidity contracts into ABI and BIN. This requires solc somewhere in the $PATH
        and also the :ref:`ethereum.tools` python library.  The return value is a dict that
        should be written into contracts.json.
        """
        ret = {}
        old_working_dir = Path.cwd()
        chdir(_BASE)

        def relativise(path: Path) -> Path:
            return path.relative_to(_BASE)

        import_dir_map = [
            "%s=%s" % (k, relativise(v)) for k, v in self.contracts_source_dirs.items()
        ]
        import_dir_map.insert(0, ".=.")  # allow solc to compile contracts in all subdirs
        try:
            for contracts_dir in self.contracts_source_dirs.values():
                res = compile_files(
                    [str(relativise(file)) for file in contracts_dir.glob("*.sol")],
                    output_values=PRECOMPILED_DATA_FIELDS + ["ast"],
                    import_remappings=import_dir_map,
                    optimize=False,
                )

                # Strip `ast` part from result
                # TODO: Remove after https://github.com/ethereum/py-solc/issues/56 is fixed
                res = {
                    contract_name: {
                        content_key: content_value
                        for content_key, content_value in contract_content.items()
                        if content_key != "ast"
                    }
                    for contract_name, contract_content in res.items()
                }
                ret.update(_fix_contract_key_names(res))
        except FileNotFoundError as ex:
            raise ContractSourceManagerCompilationError(
                "Could not compile the contract. Check that solc is available."
            ) from ex
        finally:
            chdir(old_working_dir)
        check_runtime_codesize(ret)
        return ret

    def compile_contracts(self, target_path: Path) -> ContractManager:
        """ Store compiled contracts JSON at `target_path`. """
        self.checksum_contracts()

        if self.overall_checksum is None:
            raise ContractSourceManagerCompilationError("Checksumming failed.")

        contracts_compiled = self._compile_all_contracts()

        target_path.parent.mkdir(parents=True, exist_ok=True)
        with target_path.open(mode="w") as target_file:
            target_file.write(
                json.dumps(
                    dict(
                        contracts=contracts_compiled,
                        contracts_checksums=self.contracts_checksums,
                        overall_checksum=self.overall_checksum,
                        contracts_version=None,
                    ),
                    sort_keys=True,
                    indent=4,
                )
            )

        return ContractManager(target_path)

    def verify_precompiled_checksums(self, precompiled_path: Path) -> None:
        """ Compare source code checksums with those from a precompiled file

        Throws AttributeError if called before checksum_contracts(). """

        # We get the precompiled file data
        contracts_precompiled = ContractManager(precompiled_path)

        # Compare each contract source code checksum with the one from the precompiled file
        for contract, checksum in self.contracts_checksums.items():
            try:
                # Silence mypy
                assert contracts_precompiled.contracts_checksums is not None
                precompiled_checksum = contracts_precompiled.contracts_checksums[contract]
            except KeyError:
                raise ContractSourceManagerVerificationError(f"No checksum for {contract}")
            if precompiled_checksum != checksum:
                raise ContractSourceManagerVerificationError(
                    f"checksum of {contract} does not match {precompiled_checksum} != {checksum}"
                )

        # Compare the overall source code checksum with the one from the precompiled file
        if self.overall_checksum != contracts_precompiled.overall_checksum:
            raise ContractSourceManagerVerificationError(
                f"overall checksum does not match "
                f"{self.overall_checksum} != {contracts_precompiled.overall_checksum}"
            )

    def checksum_contracts(self) -> None:
        """Remember the checksum of each source, and the overall checksum."""
        checksums: Dict[str, str] = {}
        for contracts_dir in self.contracts_source_dirs.values():
            file: Path
            for file in contracts_dir.glob("*.sol"):
                checksums[file.name] = hashlib.sha256(file.read_bytes()).hexdigest()

        self.overall_checksum = hashlib.sha256(
            ":".join(checksums[key] for key in sorted(checksums)).encode()
        ).hexdigest()
        self.contracts_checksums = checksums


def contracts_source_path() -> Dict[str, Path]:
    return contracts_source_path_with_stem(Path("data/source"))


def contracts_source_path_of_deployment_module(module: DeploymentModule) -> Path:
    if module == DeploymentModule.RAIDEN:
        return contracts_source_path()["raiden"]
    elif module == DeploymentModule.SERVICES:
        return contracts_source_path()["services"]
    else:
        raise ValueError(f"No source known for module {module}")


def contracts_source_path_with_stem(stem: Path) -> Dict[str, Path]:
    """The directory remapping given to the Solidity compiler."""
    return {
        "lib": _BASE.joinpath(stem, "lib"),
        "raiden": _BASE.joinpath(stem, "raiden"),
        "test": _BASE.joinpath(stem, "test"),
        "services": _BASE.joinpath(stem, "services"),
    }


def _fix_contract_key_names(d: Dict) -> Dict:
    result = {}

    for k, v in d.items():
        name = k.split(":")[1]
        result[name] = v

    return result


def check_runtime_codesize(d: Dict) -> None:
    """ Raises a RuntimeError if the runtime codesize exceeds the EIP-170 limit """
    for name, compilation in d.items():
        runtime_code_len = len(compilation["bin-runtime"]) // 2
        if runtime_code_len > 0x6000:
            raise RuntimeError(f"{name}'s runtime code is too big ({runtime_code_len} bytes).")



from typing import Any, Callable, Collection, List

import pytest
from eth_typing import HexAddress
from eth_utils import is_address
from web3 import Web3
from web3.contract import Contract, get_event_data

from raiden_contracts.constants import (
    CONTRACT_TOKEN_NETWORK,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    EVENT_TOKEN_NETWORK_CREATED,
    TEST_SETTLE_TIMEOUT_MAX,
    TEST_SETTLE_TIMEOUT_MIN,
)
from raiden_contracts.contract_manager import ContractManager
from raiden_contracts.tests.utils.constants import CONTRACT_DEPLOYER_ADDRESS
from raiden_contracts.utils.transaction import check_successful_tx


@pytest.fixture()
def get_token_network_registry(deploy_tester_contract: Contract) -> Callable:
    def get(arguments: Collection[Any]) -> Contract:
        return deploy_tester_contract(CONTRACT_TOKEN_NETWORK_REGISTRY, arguments)

    return get


@pytest.fixture(scope="session")
def token_network_registry_constructor_args(
    web3: Web3, secret_registry_contract: Contract
) -> List:
    return [
        secret_registry_contract.address,
        int(web3.version.network),
        TEST_SETTLE_TIMEOUT_MIN,
        TEST_SETTLE_TIMEOUT_MAX,
        1,
    ]


@pytest.fixture(scope="session")
def token_network_registry_contract(
    deploy_tester_contract: Callable, token_network_registry_constructor_args: List
) -> Contract:
    """Deployed TokenNetworkRegistry contract"""
    return deploy_tester_contract(
        CONTRACT_TOKEN_NETWORK_REGISTRY, token_network_registry_constructor_args
    )


@pytest.fixture
def token_network_registry_address(token_network_registry_contract: Contract) -> HexAddress:
    """Address of TokenNetworkRegistry contract"""
    return token_network_registry_contract.address


@pytest.fixture
def add_and_register_token(
    web3: Web3,
    token_network_registry_contract: Contract,
    deploy_token_contract: Contract,
    contracts_manager: ContractManager,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
) -> Callable:
    """Deploy a token and register it in TokenNetworkRegistry"""

    def f(initial_amount: int, decimals: int, token_name: str, token_symbol: str) -> Contract:
        token_contract = deploy_token_contract(initial_amount, decimals, token_name, token_symbol)
        txid = token_network_registry_contract.functions.createERC20TokenNetwork(
            token_contract.address, channel_participant_deposit_limit, token_network_deposit_limit
        ).call_and_transact({"from": CONTRACT_DEPLOYER_ADDRESS})
        (tx_receipt, _) = check_successful_tx(web3, txid)
        assert len(tx_receipt["logs"]) == 1
        event_abi = contracts_manager.get_event_abi(
            CONTRACT_TOKEN_NETWORK_REGISTRY, EVENT_TOKEN_NETWORK_CREATED
        )
        decoded_event = get_event_data(event_abi, tx_receipt["logs"][0])
        assert decoded_event is not None
        assert is_address(decoded_event["args"]["token_address"])
        assert is_address(decoded_event["args"]["token_network_address"])
        token_network_address = decoded_event["args"]["token_network_address"]
        token_network_abi = contracts_manager.get_contract_abi(CONTRACT_TOKEN_NETWORK)
        return web3.eth.contract(abi=token_network_abi, address=token_network_address)

    return f



from typing import Callable, Tuple

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import UserDepositEvent


def test_deposit(
    user_deposit_contract: Contract, custom_token: Contract, get_accounts: Callable[[int], Tuple]
) -> None:
    (A, B) = get_accounts(2)
    custom_token.functions.mint(100).call_and_transact({"from": A})
    custom_token.functions.approve(user_deposit_contract.address, 30).call_and_transact(
        {"from": A}
    )

    # deposit to A's own balance
    user_deposit_contract.functions.deposit(A, 10).call_and_transact({"from": A})
    assert user_deposit_contract.functions.balances(A).call() == 10
    assert user_deposit_contract.functions.total_deposit(A).call() == 10
    assert custom_token.functions.balanceOf(A).call() == 90
    assert custom_token.functions.balanceOf(user_deposit_contract.address).call() == 10

    # increase A's deposit
    user_deposit_contract.functions.deposit(A, 20).call_and_transact({"from": A})
    assert user_deposit_contract.functions.balances(A).call() == 20
    assert user_deposit_contract.functions.total_deposit(A).call() == 20
    assert custom_token.functions.balanceOf(A).call() == 80
    assert custom_token.functions.balanceOf(user_deposit_contract.address).call() == 20

    # a deposit can't be decreased by calling deposit
    with pytest.raises(TransactionFailed):
        user_deposit_contract.functions.deposit(A, 19).call({"from": A})

    # A deposits to the benefit of B
    user_deposit_contract.functions.deposit(B, 10).call_and_transact({"from": A})
    assert user_deposit_contract.functions.balances(B).call() == 10
    assert user_deposit_contract.functions.total_deposit(B).call() == 10
    assert custom_token.functions.balanceOf(A).call() == 70
    assert custom_token.functions.balanceOf(user_deposit_contract.address).call() == 30

    # Can't deposit more than the token contract allows
    with pytest.raises(TransactionFailed):
        user_deposit_contract.functions.deposit(A, 21).call({"from": A})

    # Can't deposit more than the whole_balance_limit
    limit = user_deposit_contract.functions.whole_balance_limit().call()
    assert limit > 0
    custom_token.functions.mint(limit + 1).call_and_transact({"from": A})
    custom_token.functions.approve(user_deposit_contract.address, limit + 1).call_and_transact(
        {"from": A}
    )
    with pytest.raises(TransactionFailed):
        user_deposit_contract.functions.deposit(A, limit + 1).call({"from": A})


def test_transfer(
    uninitialized_user_deposit_contract: Contract,
    udc_transfer_contract: Contract,
    get_accounts: Callable[[int], Tuple],
    event_handler: Callable,
    custom_token: Contract,
) -> None:
    user_deposit_contract = uninitialized_user_deposit_contract
    ev_handler = event_handler(user_deposit_contract)
    (A, B) = get_accounts(2)
    custom_token.functions.mint(10).call_and_transact({"from": A})
    custom_token.functions.approve(user_deposit_contract.address, 10).call_and_transact(
        {"from": A}
    )
    user_deposit_contract.functions.deposit(A, 10).call_and_transact({"from": A})

    # only trusted contracts can call transfer (init has not been called, yet)
    with pytest.raises(TransactionFailed):
        udc_transfer_contract.functions.transfer(A, B, 10).call()

    # happy case
    user_deposit_contract.functions.init(
        udc_transfer_contract.address, udc_transfer_contract.address
    ).call_and_transact()
    tx_hash = udc_transfer_contract.functions.transfer(A, B, 10).call_and_transact()
    ev_handler.assert_event(tx_hash, UserDepositEvent.BALANCE_REDUCED, dict(owner=A, newBalance=0))
    assert user_deposit_contract.functions.balances(A).call() == 0
    assert user_deposit_contract.functions.balances(B).call() == 10

    # no tokens left
    assert not udc_transfer_contract.functions.transfer(A, B, 1).call()

    # (not) enough tokens left
    assert udc_transfer_contract.functions.transfer(B, A, 10).call()
    assert not udc_transfer_contract.functions.transfer(B, A, 11).call()


def test_deposit_after_transfer(
    uninitialized_user_deposit_contract: Contract,
    udc_transfer_contract: Contract,
    custom_token: Contract,
    get_accounts: Callable[[int], Tuple],
) -> None:
    """ Make sure that `total_deposit` and `balance` are not mixed up.

    When doing a deposit followed by a transfer, both variables start to differ
    and we can use another deposit to verify that each is handled correctly.
    """
    user_deposit_contract = uninitialized_user_deposit_contract
    user_deposit_contract.functions.init(
        udc_transfer_contract.address, udc_transfer_contract.address
    ).call_and_transact()
    (A, B) = get_accounts(2)
    custom_token.functions.mint(100).call_and_transact({"from": A})
    custom_token.functions.approve(user_deposit_contract.address, 30).call_and_transact(
        {"from": A}
    )

    # deposit + transact
    user_deposit_contract.functions.deposit(A, 10).call_and_transact({"from": A})
    udc_transfer_contract.functions.transfer(A, B, 10).call_and_transact()
    assert user_deposit_contract.functions.balances(A).call() == 0
    assert user_deposit_contract.functions.total_deposit(A).call() == 10
    assert custom_token.functions.balanceOf(A).call() == 90
    assert custom_token.functions.balanceOf(user_deposit_contract.address).call() == 10

    # check after another deposit
    user_deposit_contract.functions.deposit(A, 20).call_and_transact({"from": A})
    assert user_deposit_contract.functions.balances(A).call() == 10
    assert user_deposit_contract.functions.total_deposit(A).call() == 20
    assert custom_token.functions.balanceOf(A).call() == 80
    assert custom_token.functions.balanceOf(user_deposit_contract.address).call() == 20


def test_withdraw(
    user_deposit_contract: Contract,
    deposit_to_udc: Callable,
    get_accounts: Callable,
    web3: Web3,
    event_handler: Callable,
) -> None:
    """ Test the interaction between planWithdraw, widthdraw and effectiveBalance
    """
    ev_handler = event_handler(user_deposit_contract)
    (A,) = get_accounts(1)
    deposit_to_udc(A, 30)
    assert user_deposit_contract.functions.balances(A).call() == 30
    assert user_deposit_contract.functions.effectiveBalance(A).call() == 30

    # plan withdraw of 20 tokens
    tx_hash = user_deposit_contract.functions.planWithdraw(20).call_and_transact({"from": A})
    ev_handler.assert_event(
        tx_hash, UserDepositEvent.WITHDRAW_PLANNED, dict(withdrawer=A, plannedBalance=10)
    )
    assert user_deposit_contract.functions.balances(A).call() == 30
    assert user_deposit_contract.functions.effectiveBalance(A).call() == 10

    # withdraw won't work before withdraw_delay elapsed
    withdraw_delay = user_deposit_contract.functions.withdraw_delay().call()
    web3.testing.mine(withdraw_delay - 1)
    with pytest.raises(TransactionFailed):
        user_deposit_contract.functions.withdraw(18).call({"from": A})

    # can't withdraw more then planned
    web3.testing.mine(1)  # now withdraw_delay is over
    with pytest.raises(TransactionFailed):
        user_deposit_contract.functions.withdraw(21).call({"from": A})

    # actually withdraw 18 tokens
    user_deposit_contract.functions.withdraw(18).call_and_transact({"from": A})
    assert user_deposit_contract.functions.balances(A).call() == 12
    assert user_deposit_contract.functions.effectiveBalance(A).call() == 12



from typing import Callable

import pytest
from web3.contract import Contract

from raiden_contracts.constants import CONTRACT_SERVICE_REGISTRY, EMPTY_ADDRESS
from raiden_contracts.tests.utils import (
    CONTRACT_DEPLOYER_ADDRESS,
    DEFAULT_BUMP_DENOMINATOR,
    DEFAULT_BUMP_NUMERATOR,
    DEFAULT_DECAY_CONSTANT,
    DEFAULT_MIN_PRICE,
    DEFAULT_REGISTRATION_DURATION,
)


@pytest.fixture(scope="session")
def service_registry(deploy_tester_contract: Callable, custom_token: Contract) -> Contract:
    return deploy_tester_contract(
        CONTRACT_SERVICE_REGISTRY,
        [
            custom_token.address,
            CONTRACT_DEPLOYER_ADDRESS,
            int(3000e18),
            DEFAULT_BUMP_NUMERATOR,
            DEFAULT_BUMP_DENOMINATOR,
            DEFAULT_DECAY_CONSTANT,
            DEFAULT_MIN_PRICE,
            DEFAULT_REGISTRATION_DURATION,
        ],
    )


@pytest.fixture(scope="session")
def service_registry_without_controller(
    deploy_tester_contract: Callable, custom_token: Contract
) -> Contract:
    return deploy_tester_contract(
        CONTRACT_SERVICE_REGISTRY,
        [
            custom_token.address,
            EMPTY_ADDRESS,
            3000 * (10 ** 18),
            DEFAULT_BUMP_NUMERATOR,
            DEFAULT_BUMP_DENOMINATOR,
            DEFAULT_DECAY_CONSTANT,
            DEFAULT_MIN_PRICE,
            DEFAULT_REGISTRATION_DURATION,
        ],
    )



import random

from raiden_contracts.tests.utils.constants import UINT256_MAX
from raiden_contracts.utils.signature import private_key_to_address


def get_random_privkey() -> str:
    """Returns a random private key"""
    return "0x%064x" % random.randint(1, UINT256_MAX)


def get_random_address() -> str:
    """Returns a random valid ethereum address"""
    return private_key_to_address(get_random_privkey())



from typing import Callable, List, Tuple

import pytest
from web3.contract import Contract

from raiden_contracts.constants import CONTRACT_CUSTOM_TOKEN, CONTRACT_HUMAN_STANDARD_TOKEN

CUSTOM_TOKEN_TOTAL_SUPPLY = 10 ** 26


@pytest.fixture(scope="session")
def token_args() -> Tuple:
    return (CUSTOM_TOKEN_TOTAL_SUPPLY, 18, CONTRACT_CUSTOM_TOKEN, "TKN")


@pytest.fixture(scope="session")
def custom_token_factory(deploy_tester_contract: Callable, token_args: List) -> Callable:
    """A function that deploys a CustomToken contract"""

    def f() -> Contract:
        return deploy_tester_contract(CONTRACT_CUSTOM_TOKEN, token_args)

    return f


@pytest.fixture(scope="session")
def custom_token(custom_token_factory: Callable) -> Contract:
    """Deploy CustomToken contract"""
    return custom_token_factory()


@pytest.fixture()
def human_standard_token(deploy_token_contract: Callable, token_args: List) -> Contract:
    """Deploy HumanStandardToken contract"""
    return deploy_token_contract(*token_args)


@pytest.fixture
def deploy_token_contract(deploy_tester_contract: Contract) -> Callable:
    """Returns a function that deploys a generic HumanStandardToken contract"""

    def f(initial_amount: int, decimals: int, token_name: str, token_symbol: str) -> Contract:
        assert initial_amount > 0
        assert decimals > 0
        return deploy_tester_contract(
            CONTRACT_HUMAN_STANDARD_TOKEN, [initial_amount, decimals, token_name, token_symbol]
        )

    return f


@pytest.fixture
def standard_token_contract(custom_token: Contract) -> Contract:
    """Deployed CustomToken contract"""
    return custom_token



from typing import Callable, List, Optional

import pytest
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import (
    CONTRACT_MONITORING_SERVICE,
    CONTRACT_ONE_TO_N,
    CONTRACT_SECRET_REGISTRY,
    CONTRACT_SERVICE_REGISTRY,
    CONTRACT_TOKEN_NETWORK,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    CONTRACT_USER_DEPOSIT,
    TEST_SETTLE_TIMEOUT_MAX,
    TEST_SETTLE_TIMEOUT_MIN,
)
from raiden_contracts.contract_manager import gas_measurements
from raiden_contracts.tests.utils.constants import (
    CONTRACT_DEPLOYER_ADDRESS,
    SERVICE_DEPOSIT,
    UINT256_MAX,
)
from raiden_contracts.utils.pending_transfers import get_locked_amount, get_pending_transfers_tree
from raiden_contracts.utils.proofs import sign_one_to_n_iou, sign_reward_proof


@pytest.mark.parametrize("version", [None])
def test_gas_json_has_enough_fields(version: Optional[str]) -> None:
    """ Check is gas.json contains enough fields """
    doc = gas_measurements(version)
    keys = {
        "CustomToken.mint",
        "CustomToken.approve",
        "CustomToken.transfer",
        "CustomToken.transferFrom",
        "MonitoringService.claimReward",
        "MonitoringService.monitor",
        "OneToN.claim",
        "SecretRegistry.registerSecret",
        "ServiceRegistry.deposit",
        "ServiceRegistry.setURL",
        "TokenNetwork DEPLOYMENT",
        "TokenNetwork.closeChannel",
        "TokenNetwork.openChannel",
        "TokenNetwork.setTotalDeposit",
        "TokenNetwork.setTotalWithdraw",
        "TokenNetwork.settleChannel",
        "TokenNetwork.unlock 1 locks",
        "TokenNetwork.unlock 6 locks",
        "TokenNetwork.updateNonClosingBalanceProof",
        "TokenNetworkRegistry DEPLOYMENT",
        "TokenNetworkRegistry createERC20TokenNetwork",
        "UserDeposit.deposit",
        "UserDeposit.deposit (increase balance)",
        "UserDeposit.planWithdraw",
        "UserDeposit.withdraw",
    }
    assert set(doc.keys()) == keys


@pytest.fixture
def print_gas_token_network_registry(
    web3: Web3,
    deploy_tester_contract_txhash: Callable,
    secret_registry_contract: Contract,
    print_gas: Callable,
) -> None:
    """ Abusing pytest to print the deployment gas cost of TokenNetworkRegistry """
    txhash = deploy_tester_contract_txhash(
        CONTRACT_TOKEN_NETWORK_REGISTRY,
        [
            secret_registry_contract.address,
            int(web3.version.network),
            TEST_SETTLE_TIMEOUT_MIN,
            TEST_SETTLE_TIMEOUT_MAX,
            10,
        ],
    )
    print_gas(txhash, CONTRACT_TOKEN_NETWORK_REGISTRY + " DEPLOYMENT")


@pytest.fixture
def print_gas_token_network_deployment(
    web3: Web3,
    get_accounts: Callable,
    print_gas: Callable,
    custom_token: Contract,
    secret_registry_contract: Contract,
    deploy_tester_contract_txhash: Callable,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
) -> None:
    """ Abusing pytest to print the deployment gas cost of TokenNetwork """
    deprecation_executor = get_accounts(1)[0]
    txhash = deploy_tester_contract_txhash(
        CONTRACT_TOKEN_NETWORK,
        [
            custom_token.address,
            secret_registry_contract.address,
            int(web3.version.network),
            TEST_SETTLE_TIMEOUT_MIN,
            TEST_SETTLE_TIMEOUT_MAX,
            deprecation_executor,
            channel_participant_deposit_limit,
            token_network_deposit_limit,
        ],
    )
    print_gas(txhash, CONTRACT_TOKEN_NETWORK + " DEPLOYMENT")


@pytest.fixture
def print_gas_token_network_create(
    print_gas: Callable,
    custom_token: Contract,
    get_token_network_registry: Callable,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
    token_network_registry_constructor_args: List,
) -> None:
    """ Abusing pytest to print gas cost of TokenNetworkRegistry's createERC20TokenNetwork() """
    registry = get_token_network_registry(token_network_registry_constructor_args)
    txn_hash = registry.functions.createERC20TokenNetwork(
        custom_token.address, channel_participant_deposit_limit, token_network_deposit_limit
    ).call_and_transact({"from": CONTRACT_DEPLOYER_ADDRESS})

    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK_REGISTRY + " createERC20TokenNetwork")


@pytest.fixture
def print_gas_secret_registry(secret_registry_contract: Contract, print_gas: Callable) -> None:
    """ Abusing pytest to print gas cost of SecretRegistry's registerSecret() """
    secret = b"secretsecretsecretsecretsecretse"
    txn_hash = secret_registry_contract.functions.registerSecret(secret).call_and_transact()
    print_gas(txn_hash, CONTRACT_SECRET_REGISTRY + ".registerSecret")


@pytest.fixture
def print_gas_channel_cycle(
    web3: Web3,
    token_network: Contract,
    create_channel: Callable,
    channel_deposit: Callable,
    withdraw_channel: Callable,
    secret_registry_contract: Contract,
    get_accounts: Callable,
    print_gas: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
) -> None:
    """ Abusing pytest to print gas costs of TokenNetwork's operations """
    (A, B, C, D) = get_accounts(4)
    settle_timeout = 11

    (channel_identifier, txn_hash) = create_channel(A, B, settle_timeout)
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".openChannel")

    (_, txn_hash) = create_channel(C, D, settle_timeout)
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".openChannel")

    txn_hash = channel_deposit(channel_identifier, A, 20, B)
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".setTotalDeposit")

    txn_hash = channel_deposit(channel_identifier, B, 10, A)
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".setTotalDeposit")

    txn_hash = withdraw_channel(channel_identifier, A, 5, UINT256_MAX, B)
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".setTotalWithdraw")

    pending_transfers_tree1 = get_pending_transfers_tree(web3, [1, 1, 2, 3], [2, 1])
    locksroot1 = pending_transfers_tree1.hash_of_packed_transfers
    locked_amount1 = get_locked_amount(pending_transfers_tree1.transfers)

    pending_transfers_tree2 = get_pending_transfers_tree(web3, [3], [], 7)
    locksroot2 = pending_transfers_tree2.hash_of_packed_transfers
    locked_amount2 = get_locked_amount(pending_transfers_tree2.transfers)

    balance_proof_A = create_balance_proof(
        channel_identifier, A, 10, locked_amount1, 5, locksroot1
    )
    balance_proof_update_signature_B = create_balance_proof_countersignature(
        B, channel_identifier, *balance_proof_A
    )
    balance_proof_B = create_balance_proof(channel_identifier, B, 5, locked_amount2, 3, locksroot2)
    closing_sig_A = create_balance_proof_countersignature(A, channel_identifier, *balance_proof_B)

    for lock in pending_transfers_tree1.unlockable:
        txn_hash = secret_registry_contract.functions.registerSecret(lock[3]).call_and_transact(
            {"from": A}
        )
    print_gas(txn_hash, CONTRACT_SECRET_REGISTRY + ".registerSecret")

    for lock in pending_transfers_tree2.unlockable:
        txn_hash = secret_registry_contract.functions.registerSecret(lock[3]).call_and_transact(
            {"from": A}
        )
    print_gas(txn_hash, CONTRACT_SECRET_REGISTRY + ".registerSecret")

    txn_hash = token_network.functions.closeChannel(
        channel_identifier, B, A, *balance_proof_B, closing_sig_A
    ).call_and_transact({"from": A})
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".closeChannel")

    txn_hash = token_network.functions.updateNonClosingBalanceProof(
        channel_identifier, A, B, *balance_proof_A, balance_proof_update_signature_B
    ).call_and_transact({"from": B})
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".updateNonClosingBalanceProof")

    web3.testing.mine(settle_timeout)
    txn_hash = token_network.functions.settleChannel(
        channel_identifier, B, 5, locked_amount2, locksroot2, A, 10, locked_amount1, locksroot1
    ).call_and_transact()
    print_gas(txn_hash, CONTRACT_TOKEN_NETWORK + ".settleChannel")

    txn_hash = token_network.functions.unlock(
        channel_identifier, A, B, pending_transfers_tree2.packed_transfers
    ).call_and_transact()
    print_gas(
        txn_hash,
        "{0}.unlock {1} locks".format(
            CONTRACT_TOKEN_NETWORK, len(pending_transfers_tree2.transfers)
        ),
    )

    txn_hash = token_network.functions.unlock(
        channel_identifier, B, A, pending_transfers_tree1.packed_transfers
    ).call_and_transact()
    print_gas(
        txn_hash,
        "{0}.unlock {1} locks".format(
            CONTRACT_TOKEN_NETWORK, len(pending_transfers_tree1.transfers)
        ),
    )


@pytest.fixture
def print_gas_monitoring_service(
    token_network: Contract,
    monitoring_service_external: Contract,
    get_accounts: Callable,
    create_channel: Callable,
    create_balance_proof: Callable,
    create_balance_proof_countersignature: Callable,
    service_registry: Contract,
    custom_token: Contract,
    deposit_to_udc: Callable,
    print_gas: Callable,
    get_private_key: Callable,
) -> None:
    """ Abusing pytest to print gas cost of MonitoringService functions """
    # setup: two parties + MS
    (A, B, MS) = get_accounts(3)
    reward_amount = 10
    deposit_to_udc(B, reward_amount)

    # register MS in the ServiceRegistry contract
    custom_token.functions.mint(SERVICE_DEPOSIT * 2).call_and_transact({"from": MS})
    custom_token.functions.approve(service_registry.address, SERVICE_DEPOSIT).call_and_transact(
        {"from": MS}
    )
    service_registry.functions.deposit(SERVICE_DEPOSIT).call_and_transact({"from": MS})

    # open a channel (c1, c2)
    channel_identifier = create_channel(A, B)[0]

    # create balance and reward proofs
    balance_proof_A = create_balance_proof(channel_identifier, B, transferred_amount=10, nonce=1)
    closing_sig_A = create_balance_proof_countersignature(A, channel_identifier, *balance_proof_A)
    balance_proof_B = create_balance_proof(channel_identifier, A, transferred_amount=20, nonce=2)
    non_closing_signature_B = create_balance_proof_countersignature(
        B, channel_identifier, *balance_proof_B
    )
    reward_proof_signature = sign_reward_proof(
        privatekey=get_private_key(B),
        monitoring_service_contract_address=monitoring_service_external.address,
        chain_id=token_network.functions.chain_id().call(),
        reward_amount=reward_amount,
        non_closing_signature=non_closing_signature_B,
    )

    # c1 closes channel
    txn_hash = token_network.functions.closeChannel(
        channel_identifier, B, A, *balance_proof_A, closing_sig_A
    ).call_and_transact({"from": A})
    token_network.web3.testing.mine(4)

    # MS calls `MSC::monitor()` using c1's BP and reward proof
    txn_hash = monitoring_service_external.functions.monitor(
        A,
        B,
        balance_proof_B[0],  # balance_hash
        balance_proof_B[1],  # nonce
        balance_proof_B[2],  # additional_hash
        balance_proof_B[3],  # closing signature
        non_closing_signature_B,  # non-closing signature
        reward_amount,
        token_network.address,  # token network address
        reward_proof_signature,
    ).call_and_transact({"from": MS})
    print_gas(txn_hash, CONTRACT_MONITORING_SERVICE + ".monitor")

    token_network.web3.testing.mine(1)

    # MS claims the reward
    txn_hash = monitoring_service_external.functions.claimReward(
        channel_identifier, token_network.address, A, B
    ).call_and_transact({"from": MS})
    print_gas(txn_hash, CONTRACT_MONITORING_SERVICE + ".claimReward")


@pytest.fixture
def print_gas_one_to_n(
    one_to_n_contract: Contract,
    deposit_to_udc: Callable,
    get_accounts: Callable,
    get_private_key: Callable,
    web3: Web3,
    print_gas: Callable,
) -> None:
    """ Abusing pytest to print gas cost of OneToN functions """
    (A, B) = get_accounts(2)
    deposit_to_udc(A, 30)

    # happy case
    chain_id = int(web3.version.network)
    amount = 10
    expiration = web3.eth.blockNumber + 2
    signature = sign_one_to_n_iou(
        get_private_key(A),
        sender=A,
        receiver=B,
        amount=amount,
        expiration_block=expiration,
        one_to_n_address=one_to_n_contract.address,
        chain_id=chain_id,
    )
    txn_hash = one_to_n_contract.functions.claim(
        A, B, amount, expiration, one_to_n_contract.address, signature
    ).call_and_transact({"from": A})
    print_gas(txn_hash, CONTRACT_ONE_TO_N + ".claim")


@pytest.fixture
def print_gas_user_deposit(
    user_deposit_contract: Contract,
    custom_token: Contract,
    get_accounts: Callable,
    web3: Web3,
    print_gas: Callable,
) -> None:
    """ Abusing pytest to print gas cost of UserDeposit functions

    The `transfer` function is not included because it's only called by trusted
    contracts as part of another function.
    """
    (A,) = get_accounts(1)
    custom_token.functions.mint(20).call_and_transact({"from": A})
    custom_token.functions.approve(user_deposit_contract.address, 20).call_and_transact(
        {"from": A}
    )

    # deposit
    txn_hash = user_deposit_contract.functions.deposit(A, 10).call_and_transact({"from": A})
    print_gas(txn_hash, CONTRACT_USER_DEPOSIT + ".deposit")
    txn_hash = user_deposit_contract.functions.deposit(A, 20).call_and_transact({"from": A})
    print_gas(txn_hash, CONTRACT_USER_DEPOSIT + ".deposit (increase balance)")

    # plan withdraw
    txn_hash = user_deposit_contract.functions.planWithdraw(10).call_and_transact({"from": A})
    print_gas(txn_hash, CONTRACT_USER_DEPOSIT + ".planWithdraw")

    # withdraw
    withdraw_delay = user_deposit_contract.functions.withdraw_delay().call()
    web3.testing.mine(withdraw_delay)
    txn_hash = user_deposit_contract.functions.withdraw(10).call_and_transact({"from": A})
    print_gas(txn_hash, CONTRACT_USER_DEPOSIT + ".withdraw")


@pytest.fixture
def print_gas_service_registry(
    get_accounts: Callable, custom_token: Contract, service_registry: Contract, print_gas: Callable
) -> None:
    (A,) = get_accounts(1)
    custom_token.functions.mint(SERVICE_DEPOSIT).call_and_transact({"from": A})
    custom_token.functions.approve(service_registry.address, SERVICE_DEPOSIT).call_and_transact(
        {"from": A}
    )

    # happy path
    deposit_tx = service_registry.functions.deposit(SERVICE_DEPOSIT).call_and_transact({"from": A})
    print_gas(deposit_tx, CONTRACT_SERVICE_REGISTRY + ".deposit")
    url = "http://example.com"
    set_url_tx = service_registry.functions.setURL(url).call_and_transact({"from": A})
    print_gas(set_url_tx, CONTRACT_SERVICE_REGISTRY + ".setURL")


@pytest.fixture
def print_gas_token(get_accounts: Callable, custom_token: Contract, print_gas: Callable) -> None:
    (A, B) = get_accounts(2)
    tx_hash = custom_token.functions.mint(100).call_and_transact({"from": A})
    print_gas(tx_hash, "CustomToken.mint")
    tx_hash = custom_token.functions.transfer(B, 100).call_and_transact({"from": A})
    print_gas(tx_hash, "CustomToken.transfer")
    tx_hash = custom_token.functions.approve(A, 100).call_and_transact({"from": B})
    print_gas(tx_hash, "CustomToken.approve")
    tx_hash = custom_token.functions.transferFrom(B, A, 100).call_and_transact({"from": A})
    print_gas(tx_hash, "CustomToken.transferFrom")


# All gas printing is done in a single test. Otherwise, after a parallel
# execution of multiple gas printing tests, you see a corrupted gas.json.
@pytest.mark.slow
@pytest.mark.usefixtures(
    "print_gas_token_network_registry",
    "print_gas_token_network_deployment",
    "print_gas_token_network_create",
    "print_gas_secret_registry",
    "print_gas_channel_cycle",
    "print_gas_monitoring_service",
    "print_gas_one_to_n",
    "print_gas_service_registry",
    "print_gas_user_deposit",
    "print_gas_token",
)
def test_print_gas() -> None:
    pass



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract, get_event_data

from raiden_contracts.constants import (
    CONTRACT_SERVICE_REGISTRY,
    EMPTY_ADDRESS,
    EVENT_REGISTERED_SERVICE,
)
from raiden_contracts.contract_manager import ContractManager, contracts_precompiled_path
from raiden_contracts.tests.utils.constants import (
    CONTRACT_DEPLOYER_ADDRESS,
    DEFAULT_BUMP_DENOMINATOR,
    DEFAULT_BUMP_NUMERATOR,
    DEFAULT_DECAY_CONSTANT,
    DEFAULT_MIN_PRICE,
    DEFAULT_REGISTRATION_DURATION,
    SECONDS_PER_DAY,
    SERVICE_DEPOSIT,
    UINT256_MAX,
)


def test_deposit_contract(
    get_deposit_contract: Callable, custom_token: Contract, get_accounts: Callable
) -> None:
    (A,) = get_accounts(1)
    custom_token.functions.mint(100).call_and_transact({"from": A})
    depo = get_deposit_contract([custom_token.address, 0, A])
    custom_token.functions.transfer(depo.address, 100).call_and_transact({"from": A})
    assert custom_token.functions.balanceOf(A).call() == 0
    assert custom_token.functions.balanceOf(depo.address).call() == 100
    depo.functions.withdraw(A).call_and_transact({"from": A})
    assert custom_token.functions.balanceOf(A).call() == 100
    assert custom_token.functions.balanceOf(depo.address).call() == 0


def test_deposit_contract_too_early_withdraw(
    get_deposit_contract: Callable, custom_token: Contract, get_accounts: Callable
) -> None:
    (A,) = get_accounts(1)
    custom_token.functions.mint(100).call_and_transact({"from": A})
    depo = get_deposit_contract([custom_token.address, UINT256_MAX, A])
    custom_token.functions.transfer(depo.address, 100).call_and_transact({"from": A})
    assert custom_token.functions.balanceOf(A).call() == 0
    assert custom_token.functions.balanceOf(depo.address).call() == 100
    with pytest.raises(TransactionFailed):
        depo.functions.withdraw(A).call_and_transact({"from": A})
    assert custom_token.functions.balanceOf(A).call() == 0
    assert custom_token.functions.balanceOf(depo.address).call() == 100


def test_deposit(
    service_registry: Contract, custom_token: Contract, get_accounts: Callable
) -> None:
    (A,) = get_accounts(1)
    custom_token.functions.mint(SERVICE_DEPOSIT).call_and_transact({"from": A})
    custom_token.functions.approve(service_registry.address, SERVICE_DEPOSIT).call_and_transact(
        {"from": A}
    )

    # happy path
    old_balance = custom_token.functions.balanceOf(A).call()
    old_price = service_registry.functions.currentPrice().call()
    old_len = service_registry.functions.everMadeDepositsLen().call()
    assert not service_registry.functions.hasValidRegistration(A).call()
    service_registry.functions.deposit(SERVICE_DEPOSIT).call_and_transact({"from": A})
    assert old_balance > custom_token.functions.balanceOf(A).call() > old_balance - old_price
    assert service_registry.functions.currentPrice().call() > old_price
    assert service_registry.functions.everMadeDepositsLen().call() == old_len + 1
    assert service_registry.functions.hasValidRegistration(A).call()
    first_expiration = service_registry.functions.service_valid_till(A).call()

    # custom_token does not allow transfer of more tokens
    with pytest.raises(TransactionFailed):
        service_registry.functions.deposit(1).call({"from": A})

    # More minting and approving before extending the registration
    custom_token.functions.mint(SERVICE_DEPOSIT).call_and_transact({"from": A})
    custom_token.functions.approve(service_registry.address, SERVICE_DEPOSIT).call_and_transact(
        {"from": A}
    )

    # Extending the registration
    service_registry.functions.deposit(SERVICE_DEPOSIT).call_and_transact({"from": A})
    assert service_registry.functions.hasValidRegistration(A).call()
    second_expiration = service_registry.functions.service_valid_till(A).call()
    assert second_expiration == first_expiration + DEFAULT_REGISTRATION_DURATION
    # This time, the list of addresses that have ever made deposits should not grow
    assert service_registry.functions.everMadeDepositsLen().call() == old_len + 1


def test_setURL(
    custom_token: Contract, service_registry: Contract, get_accounts: Callable, web3: Web3
) -> None:
    (A,) = get_accounts(1)
    url1 = "http://example.com"
    url2 = "http://raiden.example.com"

    custom_token.functions.mint(SERVICE_DEPOSIT).call_and_transact({"from": A})
    custom_token.functions.approve(service_registry.address, SERVICE_DEPOSIT).call_and_transact(
        {"from": A}
    )
    tx = service_registry.functions.deposit(SERVICE_DEPOSIT).call_and_transact({"from": A})
    tx_receipt = web3.eth.getTransactionReceipt(tx)
    contract_manager = ContractManager(contracts_precompiled_path(version=None))
    event_abi = contract_manager.get_event_abi(CONTRACT_SERVICE_REGISTRY, EVENT_REGISTERED_SERVICE)
    event_data = get_event_data(event_abi, tx_receipt["logs"][-1])
    assert event_data["args"]["service"] == A
    assert event_data["args"]["deposit_contract"] != EMPTY_ADDRESS

    service_registry.functions.setURL(url1).call_and_transact({"from": A})
    assert service_registry.functions.urls(A).call() == url1

    service_registry.functions.setURL(url2).call_and_transact({"from": A})
    assert service_registry.functions.urls(A).call() == url2


def test_decayed_price(service_registry: Contract) -> None:
    assert service_registry.functions.decayedPrice(100000, 0).call() == 100000

    # The minimum price is 1000
    assert service_registry.functions.decayedPrice(100, 0).call() == 1000

    # roughly 139 days till the price halves.
    assert service_registry.functions.decayedPrice(100000, 11990300).call() == 50000


def test_changing_duration(
    service_registry: Contract, get_accounts: Callable, custom_token: Contract
) -> None:
    new_duration = 90 * SECONDS_PER_DAY
    service_registry.functions.changeParameters(
        _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
        _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
        _decay_constant=DEFAULT_DECAY_CONSTANT,
        _min_price=DEFAULT_MIN_PRICE,
        _registration_duration=new_duration,
    ).call_and_transact({"from": CONTRACT_DEPLOYER_ADDRESS})
    assert service_registry.functions.registration_duration().call() == new_duration
    # make sure that the duration has changed.
    (A,) = get_accounts(1)
    custom_token.functions.mint(2 * SERVICE_DEPOSIT).call_and_transact({"from": A})
    custom_token.functions.approve(
        service_registry.address, 2 * SERVICE_DEPOSIT
    ).call_and_transact({"from": A})
    service_registry.functions.deposit(SERVICE_DEPOSIT).call_and_transact({"from": A})
    first_expiration = service_registry.functions.service_valid_till(A).call()
    service_registry.functions.deposit(SERVICE_DEPOSIT).call_and_transact({"from": A})
    second_expiration = service_registry.functions.service_valid_till(A).call()
    assert second_expiration == first_expiration + new_duration


def test_changing_bump_numerator(service_registry: Contract) -> None:
    service_registry.functions.changeParameters(
        _price_bump_numerator=DEFAULT_BUMP_NUMERATOR + 1,
        _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
        _decay_constant=DEFAULT_DECAY_CONSTANT,
        _min_price=DEFAULT_MIN_PRICE,
        _registration_duration=DEFAULT_REGISTRATION_DURATION,
    ).call_and_transact({"from": CONTRACT_DEPLOYER_ADDRESS})
    assert service_registry.functions.price_bump_numerator().call() == DEFAULT_BUMP_NUMERATOR + 1


def test_too_high_bump_numerator_fail(service_registry: Contract) -> None:
    """changeParameters() fails if the numerator is too big"""
    with pytest.raises(TransactionFailed):
        service_registry.functions.changeParameters(
            _price_bump_numerator=2 ** 40,
            _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
            _decay_constant=DEFAULT_DECAY_CONSTANT,
            _min_price=DEFAULT_MIN_PRICE,
            _registration_duration=DEFAULT_REGISTRATION_DURATION,
        ).call_and_transact({"from": CONTRACT_DEPLOYER_ADDRESS})


def test_changing_bump_denominator(service_registry: Contract) -> None:
    service_registry.functions.changeParameters(
        _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
        _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR + 1,
        _decay_constant=DEFAULT_DECAY_CONSTANT,
        _min_price=DEFAULT_MIN_PRICE,
        _registration_duration=DEFAULT_REGISTRATION_DURATION,
    ).call_and_transact({"from": CONTRACT_DEPLOYER_ADDRESS})
    assert (
        service_registry.functions.price_bump_denominator().call() == DEFAULT_BUMP_DENOMINATOR + 1
    )


def test_changing_too_low_bump_parameter_fail(service_registry: Contract) -> None:
    """changeParameters() fails if the bump numerator is smaller than the bump denominator"""
    with pytest.raises(TransactionFailed):
        service_registry.functions.changeParameters(
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
            _price_bump_denominator=DEFAULT_BUMP_NUMERATOR + 1,
            _decay_constant=DEFAULT_DECAY_CONSTANT,
            _min_price=DEFAULT_MIN_PRICE,
            _registration_duration=DEFAULT_REGISTRATION_DURATION,
        ).call_and_transact({"from": CONTRACT_DEPLOYER_ADDRESS})


def test_changing_decay_constant(service_registry: Contract) -> None:
    service_registry.functions.changeParameters(
        _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
        _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
        _decay_constant=DEFAULT_DECAY_CONSTANT + 100,
        _min_price=DEFAULT_MIN_PRICE,
        _registration_duration=DEFAULT_REGISTRATION_DURATION,
    ).call_and_transact({"from": CONTRACT_DEPLOYER_ADDRESS})
    assert service_registry.functions.decay_constant().call() == DEFAULT_DECAY_CONSTANT + 100


def test_too_high_decay_cosntant_fail(service_registry: Contract) -> None:
    """changeParameters() fails if the new decay constant is too high"""
    with pytest.raises(TransactionFailed):
        service_registry.functions.changeParameters(
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
            _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
            _decay_constant=2 ** 40,
            _min_price=DEFAULT_MIN_PRICE,
            _registration_duration=DEFAULT_REGISTRATION_DURATION,
        ).call_and_transact({"from": CONTRACT_DEPLOYER_ADDRESS})


def test_changing_min_price(service_registry: Contract) -> None:
    service_registry.functions.changeParameters(
        _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
        _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
        _decay_constant=DEFAULT_DECAY_CONSTANT,
        _min_price=DEFAULT_MIN_PRICE * 2,
        _registration_duration=DEFAULT_REGISTRATION_DURATION,
    ).call_and_transact({"from": CONTRACT_DEPLOYER_ADDRESS})
    assert service_registry.functions.min_price().call() == DEFAULT_MIN_PRICE * 2


def test_unauthorized_parameter_change(service_registry: Contract, get_accounts: Callable) -> None:
    """A random address's changeParameters() call should fail"""
    (A,) = get_accounts(1)
    with pytest.raises(TransactionFailed):
        service_registry.functions.changeParameters(
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
            _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
            _decay_constant=DEFAULT_DECAY_CONSTANT,
            _min_price=DEFAULT_MIN_PRICE * 2,
            _registration_duration=DEFAULT_REGISTRATION_DURATION,
        ).call_and_transact({"from": A})
    assert service_registry.functions.min_price().call() == DEFAULT_MIN_PRICE


def test_parameter_change_on_no_controller(service_registry_without_controller: Contract) -> None:
    """A random address's changeParameters() call should fail"""
    with pytest.raises(TransactionFailed):
        service_registry_without_controller.functions.changeParameters(
            _price_bump_numerator=DEFAULT_BUMP_NUMERATOR,
            _price_bump_denominator=DEFAULT_BUMP_DENOMINATOR,
            _decay_constant=DEFAULT_DECAY_CONSTANT,
            _min_price=DEFAULT_MIN_PRICE * 2,
            _registration_duration=DEFAULT_REGISTRATION_DURATION,
        ).call_and_transact({"from": CONTRACT_DEPLOYER_ADDRESS})
    assert service_registry_without_controller.functions.min_price().call() == DEFAULT_MIN_PRICE


def service_registry_with_high_numbers(
    deploy_tester_contract: Callable, custom_token: Contract
) -> Contract:
    """See if the computation overflows with the highest allowed numbers"""
    contract = deploy_tester_contract(
        CONTRACT_SERVICE_REGISTRY,
        [
            custom_token.address,
            EMPTY_ADDRESS,
            2 ** 90,
            2 ** 40 - 1,
            1,
            2 ** 40 - 1,
            DEFAULT_MIN_PRICE,
            DEFAULT_REGISTRATION_DURATION,
        ],
    )
    assert contract.functions.currentPrice() == 2 ** 90


def test_deprecation_switch(
    service_registry: Contract, get_accounts: Callable, custom_token: Contract
) -> None:
    """The controller turns on the deprecation switch and somebody tries to deposit"""
    # The controller turns on the deprecation switch
    assert not service_registry.functions.deprecation_switch().call()
    service_registry.functions.setDeprecationSwitch().call_and_transact(
        {"from": CONTRACT_DEPLOYER_ADDRESS}
    )
    assert service_registry.functions.deprecation_switch().call()
    # A user tries to make a deposit
    (A,) = get_accounts(1)
    custom_token.functions.mint(SERVICE_DEPOSIT).call_and_transact({"from": A})
    custom_token.functions.approve(service_registry.address, SERVICE_DEPOSIT).call_and_transact(
        {"from": A}
    )
    with pytest.raises(TransactionFailed):
        service_registry.functions.deposit(SERVICE_DEPOSIT).call_and_transact({"from": A})


def test_unauthorized_deprecation_switch(
    service_registry: Contract, get_accounts: Callable
) -> None:
    """A random account cannot turn on the deprecation switch"""
    (A,) = get_accounts(1)
    with pytest.raises(TransactionFailed):
        service_registry.functions.setDeprecationSwitch().call_and_transact({"from": A})



import json
import pprint
import subprocess
from pathlib import Path
from time import sleep
from typing import Dict, Optional

import click
import requests
from eth_abi import encode_abi

from raiden_contracts.constants import CONTRACT_LIST, DeploymentModule
from raiden_contracts.contract_manager import (
    ContractManager,
    DeployedContract,
    DeployedContracts,
    contracts_precompiled_path,
    get_contracts_deployment_info,
)
from raiden_contracts.contract_source_manager import (
    contracts_source_path,
    contracts_source_path_of_deployment_module,
)
from raiden_contracts.utils.type_aliases import ChainID

CONTRACT_NAMES_SEPARATED = " | ".join([c.name for c in CONTRACT_LIST])


@click.command()
@click.option("--chain-id", default=3, help="Chain id. E.g. --chain-id 3")
@click.option("--apikey", required=True, help="A valid Etherscan APIKEY is required.")
@click.option(
    "--guid", help="GUID from a previous verification attempt. Tries to get the submission status."
)
@click.option(
    "--contract-name",
    default=None,
    help=f"Contract name. Options: {CONTRACT_NAMES_SEPARATED}. "
    " Default is to submit the sources of all contracts.",
)
def etherscan_verify(
    chain_id: ChainID, apikey: str, guid: Optional[str], contract_name: Optional[str]
) -> None:
    if guid:
        guid_status(etherscan_api=api_of_chain_id[chain_id], guid=guid)
        return

    for list_entry in CONTRACT_LIST:
        if contract_name is None or contract_name == list_entry.name:
            etherscan_verify_contract(
                chain_id=chain_id,
                apikey=apikey,
                source_module=list_entry.module,
                contract_name=list_entry.name,
            )


api_of_chain_id = {
    1: "https://api.etherscan.io/api",
    3: "https://api-ropsten.etherscan.io/api",
    4: "https://api-rinkeby.etherscan.io/api",
    5: "https://api-goerli.etherscan.io/api",
    42: "https://api-kovan.etherscan.io/api",
}


def join_sources(source_module: DeploymentModule, contract_name: str) -> str:
    """ Use join-contracts.py to concatenate all imported Solidity files.

    Args:
        source_module: a module name to look up contracts_source_path()
        contract_name: 'TokenNetworkRegistry', 'SecretRegistry' etc.
    """
    joined_file = Path(__file__).parent.joinpath("joined.sol")
    remapping = {module: str(path) for module, path in contracts_source_path().items()}
    command = [
        "./utils/join-contracts.py",
        "--import-map",
        json.dumps(remapping),
        str(
            contracts_source_path_of_deployment_module(source_module).joinpath(
                contract_name + ".sol"
            )
        ),
        str(joined_file),
    ]
    working_dir = Path(__file__).parent.parent
    try:
        subprocess.check_call(command, cwd=working_dir)
    except subprocess.CalledProcessError as ex:
        print(f"cd {str(working_dir)}; {subprocess.list2cmdline(command)} failed.")
        raise ex

    return joined_file.read_text()


def get_constructor_args(
    deployment_info: DeployedContracts, contract_name: str, contract_manager: ContractManager
) -> str:
    constructor_arguments = deployment_info["contracts"][contract_name]["constructor_arguments"]
    if constructor_arguments != []:
        constructor_types = contract_manager.get_constructor_argument_types(contract_name)
        constructor_args = encode_abi(types=constructor_types, args=constructor_arguments).hex()
    else:
        constructor_types = []
        constructor_args = ""
    print("constructor_args", constructor_arguments, constructor_types, constructor_args)
    return constructor_args


def post_data_for_etherscan_verification(
    apikey: str,
    deployment_info: DeployedContract,
    source: str,
    contract_name: str,
    metadata: Dict,
    constructor_args: str,
) -> Dict:
    data = {
        # A valid API-Key is required
        "apikey": apikey,
        # Do not change
        "module": "contract",
        # Do not change
        "action": "verifysourcecode",
        "contractaddress": deployment_info["address"],
        "sourceCode": source,
        "contractname": contract_name,
        "compilerversion": "v" + metadata["compiler"]["version"],
        # 0 = Optimization used, 1 = No Optimization
        "optimizationUsed": 0 if metadata["settings"]["optimizer"]["enabled"] is False else 1,
        "runs": metadata["settings"]["optimizer"]["runs"],
        # Typo is intentional. Etherscan does not like the correct spelling.
        "constructorArguements": constructor_args,
    }
    pprint.pprint({k: v for k, v in data.items() if k != "sourceCode"})
    return data


def etherscan_verify_contract(
    chain_id: ChainID, apikey: str, source_module: DeploymentModule, contract_name: str
) -> None:
    """ Calls Etherscan API for verifying the Solidity source of a contract.

    Args:
        chain_id: EIP-155 chain id of the Ethereum chain
        apikey: key for calling Etherscan API
        source_module: a module name to look up contracts_source_path()
        contract_name: 'TokenNetworkRegistry', 'SecretRegistry' etc.
    """
    etherscan_api = api_of_chain_id[chain_id]
    deployment_info = get_contracts_deployment_info(chain_id=chain_id, module=source_module)
    if deployment_info is None:
        raise FileNotFoundError(
            f"Deployment file not found for chain_id={chain_id} and module={source_module}"
        )
    contract_manager = ContractManager(contracts_precompiled_path())

    data = post_data_for_etherscan_verification(
        apikey=apikey,
        deployment_info=deployment_info["contracts"][contract_name],
        source=join_sources(source_module=source_module, contract_name=contract_name),
        contract_name=contract_name,
        metadata=json.loads(contract_manager.contracts[contract_name]["metadata"]),
        constructor_args=get_constructor_args(
            deployment_info=deployment_info,
            contract_name=contract_name,
            contract_manager=contract_manager,
        ),
    )
    response = requests.post(etherscan_api, data=data)
    content = json.loads(response.content.decode())
    print(content)
    print(f'Status: {content["status"]}; {content["message"]} ; GUID = {content["result"]}')

    etherscan_url = etherscan_api.replace("api-", "").replace("api", "")
    etherscan_url += "/verifyContract2?a=" + data["contractaddress"]
    manual_submission_guide = f"""Usually a manual submission to Etherscan works.
    Visit {etherscan_url}
    Use raiden_contracts/deploy/joined.sol."""

    if content["status"] != "1":
        if content["result"] == "Contract source code already verified":
            return
        else:
            raise ValueError(
                "Etherscan submission failed for an unknown reason\n" + manual_submission_guide
            )

    # submission succeeded, obtained GUID
    guid = content["result"]
    status = "0"
    retries = 10
    while status == "0" and retries > 0:
        retries -= 1
        r = guid_status(etherscan_api=etherscan_api, guid=guid)
        status = r["status"]
        if r["result"] == "Fail - Unable to verify":
            raise ValueError(manual_submission_guide)
        if r["result"] == "Pass - Verified":
            return
        print("Retrying...")
        sleep(5)
    raise TimeoutError(manual_submission_guide)


def guid_status(etherscan_api: str, guid: str) -> Dict:
    data = {"guid": guid, "module": "contract", "action": "checkverifystatus"}
    r = requests.get(etherscan_api, data)
    status_content = json.loads(r.content.decode())
    print(status_content)
    return status_content


if __name__ == "__main__":
    # pylint: disable=E1120
    etherscan_verify()



import pytest
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import CONTRACT_ONE_TO_N


@pytest.fixture(scope="session")
def one_to_n_contract(
    deploy_tester_contract: Contract, uninitialized_user_deposit_contract: Contract, web3: Web3
) -> Contract:
    chain_id = int(web3.version.network)
    return deploy_tester_contract(
        CONTRACT_ONE_TO_N, [uninitialized_user_deposit_contract.address, chain_id]
    )



from typing import Callable, List

import pytest
from eth_typing import HexAddress
from web3 import Web3
from web3.contract import Contract, get_event_data

from raiden_contracts.constants import (
    CONTRACT_TOKEN_NETWORK,
    CONTRACT_TOKEN_NETWORK_REGISTRY,
    EVENT_TOKEN_NETWORK_CREATED,
    MAX_ETH_CHANNEL_PARTICIPANT,
    MAX_ETH_TOKEN_NETWORK,
    TEST_SETTLE_TIMEOUT_MAX,
    TEST_SETTLE_TIMEOUT_MIN,
)
from raiden_contracts.contract_manager import ContractManager
from raiden_contracts.tests.utils.constants import CONTRACT_DEPLOYER_ADDRESS

snapshot_before_token_network = None


@pytest.fixture
def get_token_network(deploy_tester_contract: Callable) -> Callable:
    """Deploy a token network as a separate contract (registry is not used)"""

    def get(arguments: List) -> Contract:
        return deploy_tester_contract(CONTRACT_TOKEN_NETWORK, arguments)

    return get


@pytest.fixture(scope="session")
def register_token_network(
    web3: Web3, token_network_registry_contract: Contract, contracts_manager: ContractManager
) -> Callable:
    """Returns a function that uses token_network_registry fixture to register
    and deploy a new token network"""

    def get(
        token_address: HexAddress,
        channel_participant_deposit_limit: int,
        token_network_deposit_limit: int,
    ) -> Contract:
        tx_hash = token_network_registry_contract.functions.createERC20TokenNetwork(
            token_address, channel_participant_deposit_limit, token_network_deposit_limit
        ).call_and_transact({"from": CONTRACT_DEPLOYER_ADDRESS})
        tx_receipt = web3.eth.getTransactionReceipt(tx_hash)
        event_abi = contracts_manager.get_event_abi(
            CONTRACT_TOKEN_NETWORK_REGISTRY, EVENT_TOKEN_NETWORK_CREATED
        )
        event_data = get_event_data(event_abi, tx_receipt["logs"][0])
        contract_address = event_data["args"]["token_network_address"]
        contract = web3.eth.contract(
            abi=contracts_manager.get_contract_abi(CONTRACT_TOKEN_NETWORK),
            address=contract_address,
        )
        return contract

    return get


@pytest.fixture(scope="session")
def channel_participant_deposit_limit() -> int:
    return MAX_ETH_CHANNEL_PARTICIPANT


@pytest.fixture(scope="session")
def token_network_deposit_limit() -> int:
    return MAX_ETH_TOKEN_NETWORK


@pytest.fixture
def no_token_network(web3: Web3) -> None:
    """ Some tests must be executed before a token network gets created

    These tests should use this fixture. Otherwise a session level token
    network might already be registered.
    """
    if snapshot_before_token_network is not None:
        web3.testing.revert(snapshot_before_token_network)


@pytest.fixture(scope="session")
def token_network(
    register_token_network: Callable,
    custom_token: Contract,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
    web3: Web3,
) -> Contract:
    """Register a new token network for a custom token"""
    global snapshot_before_token_network
    snapshot_before_token_network = web3.testing.snapshot()
    return register_token_network(
        custom_token.address, channel_participant_deposit_limit, token_network_deposit_limit
    )


@pytest.fixture
def token_network_contract(
    deploy_tester_contract: Contract,
    secret_registry_contract: Contract,
    standard_token_contract: Contract,
) -> Contract:
    network_id = int(secret_registry_contract.web3.version.network)
    return deploy_tester_contract(
        CONTRACT_TOKEN_NETWORK,
        [standard_token_contract.address, secret_registry_contract.address, network_id],
    )


@pytest.fixture()
def token_network_external(
    web3: Web3,
    get_token_network: Callable,
    custom_token: Contract,
    secret_registry_contract: Contract,
    channel_participant_deposit_limit: int,
    token_network_deposit_limit: int,
) -> Contract:
    return get_token_network(
        [
            custom_token.address,
            secret_registry_contract.address,
            int(web3.version.network),
            TEST_SETTLE_TIMEOUT_MIN,
            TEST_SETTLE_TIMEOUT_MAX,
            CONTRACT_DEPLOYER_ADDRESS,
            channel_participant_deposit_limit,
            token_network_deposit_limit,
        ]
    )



from typing import Callable

import pytest
from eth_tester.exceptions import TransactionFailed
from web3 import Web3
from web3.contract import Contract

from raiden_contracts.constants import TEST_SETTLE_TIMEOUT_MAX, TEST_SETTLE_TIMEOUT_MIN
from raiden_contracts.tests.utils import LOCKSROOT_OF_NO_LOCKS, fake_bytes


def test_settle_timeout_inrange(
    token_network: Contract,
    get_accounts: Callable,
    web3: Web3,
    create_close_signature_for_no_balance_proof: Callable,
) -> None:
    """ The TokenNetwork constructor must enforce that settle timeout is in
    the valid range.

    Also asserts that the constants.py and the netting channel contract values
    are synched.
    """
    (A, B) = get_accounts(2)

    small_settle_timeout = TEST_SETTLE_TIMEOUT_MIN - 1
    large_settle_timeout = TEST_SETTLE_TIMEOUT_MAX + 1

    with pytest.raises(TransactionFailed):
        token_network.functions.openChannel(A, B, small_settle_timeout).call()

    with pytest.raises(TransactionFailed):
        token_network.functions.openChannel(A, B, large_settle_timeout).call()

    token_network.functions.openChannel(A, B, TEST_SETTLE_TIMEOUT_MIN).call_and_transact()
    channel_identifier = token_network.functions.getChannelIdentifier(A, B).call()
    (settle_block_number, _) = token_network.functions.getChannelInfo(
        channel_identifier, A, B
    ).call()

    assert settle_block_number == TEST_SETTLE_TIMEOUT_MIN

    closing_sig = create_close_signature_for_no_balance_proof(A, channel_identifier)
    token_network.functions.closeChannel(
        channel_identifier=channel_identifier,
        non_closing_participant=B,
        closing_participant=A,
        balance_hash=fake_bytes(32),
        nonce=0,
        additional_hash=fake_bytes(32),
        non_closing_signature=fake_bytes(65),
        closing_signature=closing_sig,
    ).call_and_transact({"from": A})
    web3.testing.mine(TEST_SETTLE_TIMEOUT_MIN + 1)
    token_network.functions.settleChannel(
        channel_identifier, A, 0, 0, LOCKSROOT_OF_NO_LOCKS, B, 0, 0, LOCKSROOT_OF_NO_LOCKS
    ).call_and_transact({"from": A})
    token_network.functions.openChannel(A, B, TEST_SETTLE_TIMEOUT_MAX).call_and_transact()
    channel_identifier = token_network.functions.getChannelIdentifier(A, B).call()
    (settle_block_number, _) = token_network.functions.getChannelInfo(
        channel_identifier, A, B
    ).call()

    assert settle_block_number == TEST_SETTLE_TIMEOUT_MAX



#!/usr/bin/env python3
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

import os
from typing import List

from setuptools import Command
from setuptools.command.build_py import build_py


DESCRIPTION = 'Raiden contracts library and utilities'
VERSION = '0.29.0'


def read_requirements(path: str) -> List[str]:
    assert os.path.isfile(path)
    with open(path) as requirements:
        return requirements.read().split()


def _get_single_requirement(requirements: List[str], package: str) -> List[str]:
    return [req for req in requirements if req.startswith(package)]


class BuildPyCommand(build_py):
    def run(self):
        try:
            self.run_command('verify_contracts')
        except SystemExit:
            pass
        build_py.run(self)


class VerifyContracts(Command):
    description = 'Verify that the compiled contracts have the correct source code checksum'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from raiden_contracts.contract_manager import (
            contracts_precompiled_path,
        )
        from raiden_contracts.contract_source_manager import (
            ContractSourceManager,
            contracts_source_path,
        )
        manager = ContractSourceManager(contracts_source_path())
        manager.checksum_contracts()
        manager.verify_precompiled_checksums(contracts_precompiled_path())


class CompileContracts(Command):
    description = 'Compile contracts and add ABI and checksums to a json file'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from raiden_contracts.contract_manager import contracts_precompiled_path
        from raiden_contracts.contract_source_manager import (
            ContractSourceManager,
            contracts_source_path,
        )

        contract_manager = ContractSourceManager(contracts_source_path())
        contract_manager.compile_contracts(contracts_precompiled_path())


requirements = read_requirements('requirements.txt')

config = {
    'version': VERSION,
    'scripts': [],
    'name': 'raiden-contracts',
    'author': 'Brainbot Labs Est.',
    'author_email': 'contact@brainbot.li',
    'description': DESCRIPTION,
    'url': 'https://github.com/raiden-network/raiden-contracts/',
    'license': 'MIT',
    'keywords': 'raiden ethereum blockchain',
    'install_requires': requirements,
    'setup_requires': _get_single_requirement(requirements, 'py-solc'),
    'packages': find_packages(),
    'include_package_data': True,
    'classifiers': [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    'entry_points': {
        'console_scripts': ['deploy = raiden_contracts.deploy.__main__:main'],
    },
    'cmdclass': {
        'compile_contracts': CompileContracts,
        'verify_contracts': VerifyContracts,
        'build_py': BuildPyCommand,
    },
    'zip_safe': False,
    'package_data': {"raiden_contracts": ["py.typed"]},
}

setup(**config)



from enum import IntEnum

from eth_utils.units import units

from raiden_contracts.utils.signature import private_key_to_address

UINT256_MAX = 2 ** 256 - 1
FAKE_ADDRESS = "0x03432"
EMPTY_BALANCE_HASH = b"\x00" * 32
EMPTY_ADDITIONAL_HASH = b"\x00" * 32
EMPTY_SIGNATURE = b"\x00" * 65
passphrase = "0"
FAUCET_PRIVATE_KEY = "0xaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
FAUCET_ADDRESS = private_key_to_address(FAUCET_PRIVATE_KEY)
FAUCET_ALLOWANCE = 100 * int(units["ether"])
CONTRACT_DEPLOYER_ADDRESS = FAUCET_ADDRESS
NONEXISTENT_LOCKSROOT = b"\x00" * 32
SECONDS_PER_DAY = 60 * 60 * 24

# Constants for ServiceRegistry testing
SERVICE_DEPOSIT = 5000 * (10 ** 18)
DEFAULT_BUMP_NUMERATOR = 6
DEFAULT_BUMP_DENOMINATOR = 5
DEFAULT_DECAY_CONSTANT = 200 * SECONDS_PER_DAY
DEFAULT_REGISTRATION_DURATION = 180 * SECONDS_PER_DAY
DEFAULT_MIN_PRICE = 1000


class TestLockIndex(IntEnum):
    EXPIRATION = 0
    AMOUNT = 1
    SECRETHASH = 2
    SECRET = 3



import pytest
from web3.contract import Contract

from raiden_contracts.constants import CONTRACT_MONITORING_SERVICE


@pytest.fixture(scope="session")
def monitoring_service_external(
    deploy_tester_contract: Contract,
    custom_token: Contract,
    service_registry: Contract,
    uninitialized_user_deposit_contract: Contract,
) -> Contract:
    return deploy_tester_contract(
        CONTRACT_MONITORING_SERVICE,
        [
            custom_token.address,
            service_registry.address,
            uninitialized_user_deposit_contract.address,
        ],
    )


@pytest.fixture()
def monitoring_service_internals(
    custom_token: Contract,
    service_registry: Contract,
    uninitialized_user_deposit_contract: Contract,
    deploy_tester_contract: Contract,
) -> Contract:
    return deploy_tester_contract(
        "MonitoringServiceInternalsTest",
        [
            custom_token.address,
            service_registry.address,
            uninitialized_user_deposit_contract.address,
        ],
    )



from typing import Optional

from semver import compare


def contracts_version_expects_deposit_limits(contracts_version: Optional[str]) -> bool:
    """ Answers whether TokenNetworkRegistry of the contracts_vesion needs deposit limits """
    if contracts_version is None:
        # contracts_version == None means the stock version in development.
        return True
    if contracts_version == "0.3._":
        return False
    return compare(contracts_version, "0.9.0") > -1


def contracts_version_with_max_token_networks(version: Optional[str]) -> bool:
    if version is None:
        # contracts_version == None means the stock version in development.
        return True
    if version == "0.3._":
        return False
    if version == "0.8.0_unlimited":
        return False
    return compare(version, "0.9.0") >= 0


def contracts_version_provides_services(version: Optional[str]) -> bool:
    if version is None:
        # contracts_version == None means the stock version in development.
        return True
    if version == "0.3._":
        return False
    if version == "0.8.0_unlimited":
        return True
    return compare(version, "0.8.0") >= 0


def contracts_version_has_initial_service_deposit(version: Optional[str]) -> bool:
    if version is None:
        # contracts_versoin == None means the stock version in development.
        return True
    if version == "0.3._":
        return False
    if version == "0.8.0_unlimited":
        return False
    return compare(version, "0.18.0") > 0
